#!/usr/local/bin/perl

#############################################################################
# fwrules.pl: Analyze and print a report of Firewall-1 rules and objects
#
# Copyright (C) 1999 Sean O'Neill
# Copyright (C) 1998 Stefan Moser
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# This program is based on previous work by an anonymous programmer who
# posted it to the FireWall-1 mailing list. Stephan Moser modified the program
# considerably and put it under the GPL.  He has moved on and Sean O'Neill has
# taken over the maintaince of this script.
#
# For bug reports and suggestions or if you just want to talk to me please
# contact me at sean.oneill@appnet.com
#
# Hopefully one day:
#     - Add a switch to turn on / off the addition of the graphic icons.
#     - Add a switch to turn on / off the addition of the name links.
#     - Change the logic to print the graphic icons.  I don't like all the
#       if statements ... but it works for now.
#     - Add logic to process user db definitions
#     - Add switch to specify were to get graphic icons from instead of hard
#       coding path in script
#
# KNOWN ISSUES:  All the graphic objects capable of being used within FW-1 have
#                not been fully tested as my firewall ruleset does not utilize
#                all the objects.  If you have an object in your ruleset that
#                doesn't display correctly, send me updated code or a simple
#                ruleset and object file showing the problem.
#
#
# March 2000
# v 4.2 to 5.0
#     - Wei Xu		wei_xu_@hotmail.com
#     - The major change of this version is to redirect the outputs to the
#			individual files instead of one html. This is just a by-product of my
#			two-days consulting job. Why it is released to you??? One day you
#			will find how valuable it would be. Enjoy     
#     - 1) modified programming style so that it can handle both html and plain
#			text files
#     - 2) added outputs to individual files instead of STDOUT
#     - 3) added 'split_rules' so that multiple lines are generated under one rule.
#			this works only if file name for html is not specified.
#     - 4) added 'all_objs' so that network objects and services are listed
#			regardless if they are in use or not
#     - 5) the 'split_rules' is not for html outputs since I don't think
#			it provides any values by doing this.
# October 1999
# v 4.1 to v4.2 
#     - Gianluca Rotoni
#     - Fixed 'Negated objects and services' bug. The error occurred when a
#       network object or a service was negated in a rule (operator "not in").
#       The rule was parsed all right and appeared in the output but the negated
#       object or service didn't (unless the object/service was used somewhere
#       else in the rulebase in a non-negated form).
#
#     - Sean O'Neill
#     - Fixed four bugs.
#       1) Short log objects were not being handled at all.
#       2) Disabled rules were not being handled correctly. The rule would 
#          print out OK but the HTML page wouldn't indicate that the rule was 
#          actually disabled.  The word DISABLED (hyphenated) is printed 
#          beneath the rule row number.
#       3) Fixed a problem with how the script handled rules based on Session 
#          Authentication.  Noted by John Allen.
#       4) The grahpic icon for rules with "Reject" as the action were not 
#          being handled correctly.
#
# July 10 1999 - Sean O'Neill
# v4.0 to v4.1
#     - Fixed a few bugs I missed
#     - Added logic to print &nbsp; character in the Address Translation Rules
#       section if nothing was printed in a cell.
#     - Added logic to print &nbsp; character in the Security Servers
#       section if nothing was printed in a cell for the various Welcome
#       Messages.
#
# July 1999 - Sean O'Neill - AppNet, Inc. - soneill@appnet.com
# Stefan Moser has indicated he will no longer be maintaining this script.  He
# is moving on.  
# Modifications include:
#     - Added graphical icons to ruleset table.  Makes the display look more
#       like the real thing.
#     - Modified the script logic flow some.  This was done to enable the
#       addition of the graphic icons.  Basically, the objects are read in
#       before the rules are.
#     - Added href's around the graphic icon in the rules section to the object
#       definitions in the appropriate sections.
#     - Incorporated John Schoonover, Dave Long, and my updates together
#       Because of the coloring used in my firewall definitions, I made the
#       addition of the color tags off by default.  See --color in help for
#       help on this.
#
#     - NOTE that all icons types were not added.  My firewall doesn't have
#       everything defined but it has the basics defined.
#
# From fwrules3.pl:
# June 2, 1999 - Some slight but useful modifications to Stefan Moser's
# masterful work: 
#      I have added html color tags to network and service objects to reflect
#          the colors used in the FW-1 GUI. Thanks to Doug Jacobson
#          (jacobson@pop.c-com.net) for his color chart at 
#          http://www.phoenix.net/~jacobson/pages/rgbhex.html
#      I have provided anchors and links between network and service objects
#          and references to them as members of groups. This only works if
#          the linked items are in the same file as the anchors.
#      john.schoonover@usa.net.
#
# From an email addressed to Stefan Moser:
# June 1999 - Dave Long - dave.long@unitedkingdom.ncr.com
# Modifications include:
#     - Added logic to make the script properly handle FW-1 v4.0 encryption
#       rules.
#############################################################################

#
# Used modules
#
use English;
use FileHandle;
use Getopt::Long;

#
# Defaults, Constants
#
$SCR_NAME="fwrules";
$VERSION="4.2";
$TRUE = 1;
$FALSE = 0;

$DEFAULT_OBJ_FILE = 'objects.C';
$DEFAULT_RULE_FILE = 'Stadard.W';
$ICONPATH = '../icons';

@NWO = qw(host gateway router switch network domain logical group blackbox machines_range);
@SVO = qw(group icmp other rpc tcp udp ports_range);

$rules = 0;
$nohtml = $TRUE;
$all_objs = $FALSE;

%fieldIndices =
  ('src'      => 1,
   'dst'      => 2,
   'services' => 3,
   'action'   => 4,
   'track'    => 5,
   'time'     => 6,
   'install'  => 7,
   'comments' => 8,
   'Encrypt'  => 9,
   'disabled' => 10);

%adtrFieldIndices =
  ('src_adtr'                 => 1,
   'dst_adtr'                 => 2,
   'services_adtr'            => 3,
   'src_adtr_translated'      => 4,
   'dst_adtr_translated'      => 5,
   'services_adtr_translated' => 6,
   'install'                  => 7,
   'comments'                 => 8);

%colors =
  ('black'              => '<font color="#111111">',
   'blue'               => '<font color="#3333FF">',
   'cyan'               => '<font color="#00FFFF">',
   'dark green'         => '<font color="#006600">',
   'dark orchid'        => '<font color="#CC33CC">',
   'firebrick'          => '<font color="#993333">',
   'foreground'         => '<font color="#111111">',
   'forest green'       => '<font color="#66FF99">',
   'gold'               => '<font color="#FFCC33">',
   'green'              => '<font color="#00FF00">',
   'magenta'            => '<font color="#FF99FF">',
   'medium slate blue'  => '<font color="#6666FF">',
   'navy blue'          => '<font color="#000077">',
   'orange'             => '<font color="#FF9900">',
   'red'                => '<font color="#FF0000">',
   'sienna'             => '<font color="#996600">',
   'yellow'             => '<font color="#FFFF00">');

@headers =
  ('RULE',
   'SOURCE',
   'DESTINATION',
   'SERVICES',
   'ACTION',
   'TRACK',
   'SCHEDULE',
   'INSTALL',
   'COMMENTS');

@adtrHeaders =
  ('RULE',
   'O_SOURCE',
   'O_DESTINATION',
   'O_SERVICE',
   'N_SOURCE',
   'N_DESTINATION',
   'N_SERVICE',
   'INSTALL',
   'COMMENTS');


# Parse and process options
if (!GetOptions(\%optctl, 'network-objects', 'service-objects', 'properties',
		'objects|o=s', 'rules|r=s', 'types-n|tn=s', 'types-s|ts=s',
		'all', 'help', 'version', 'with_colors', 'without_icons', 'split_rules', 'all_objs')
    || keys(%optctl) == 0 || $optctl{help} == 1 || $optctl{version} == 1 || @ARGV > 1
    || (defined($optctl{'n-types'}) && !defined($optctl{'network-objects'}))
    || (defined($optctl{'s-types'}) && !defined($optctl{'service-objects'})))
{
	if ($optctl{version} == 1)
	{
		print STDERR "$SCR_NAME version $VERSION\n";
	}
	else
	{
		usage();
	}

	exit;
}

if (defined($optctl{all}))
{
	$optctl{'network-objects'} = $optctl{'service-objects'} = '';
	$optctl{properties} = 1;
}

# Parse network object types
if (defined($optctl{'network-objects'}))
{
	if (defined($optctl{'types-n'}))
	{
		@nwoTypes = split(/,/, lc($optctl{'types-n'}));
		foreach $type (@nwoTypes)
		{
			if (!grep(/^$type$/, @NWO)) {
			die "Error: invalid network object type specified: $type\n";
			}
		}
    }
	else
	{
		@nwoTypes = @NWO;
	}
}

# Parse service object type
if (defined($optctl{'service-objects'}))
{
	if (defined($optctl{'types-s'}))
	{
		@svoTypes = split(/,/, lc($optctl{'types-s'}));
		foreach $type (@svoTypes)
		{
			if (!grep(/^$type$/, @SVO))
			{
				die "Error: invalid service object type: $type\n";
			}
		}
	}
	else
	{
		@svoTypes = @SVO;
	}
}

# Open input and output files
if (defined($optctl{'rules'}))
{
	$rules = 1;
	$rFile = $optctl{'rules'};
	if (!open(RULES, $rFile))
	{
		if ($rules == 0) {
			print STDERR "$SCR_NAME: Couldn't open rule file $rFile\n";
			exit 1;
		}
	}
}

if (defined($optctl{'objects'}))
{
	$objectsFileName = $optctl{'objects'};
}
else
{
	$objectsFileName = $DEFAULT_OBJ_FILE;
}

if (defined($optctl{'with_colors'}))
{
	$with_colors = $TRUE;
}
else
{
	$with_colors = $FALSE;
}
if (defined($optctl{'without_icons'}))
{
	$without_icons= $TRUE;
}
else
{
	$without_icons= $FALSE;
}
if (defined($optctl{'all_objs'}))
{
	$all_objs = $TRUE;
}


if (!open(OBJECTS, $objectsFileName))
{
	print STDERR "$SCR_NAME: Couldn't open objects file '$objectsFileName'\n";
	exit 1;
}

if (@ARGV > 0)
{
	openStandFile($ARGV[0]);
	$nohtml = $FALSE;
	$sph = "<\/th><th>";
	$spd = "<\/td><td>";
	$split_rules = $FALSE;
} else {
	$sph = "|";
	$spd = "|";
	$without_icons= $TRUE;
	if (defined($optctl{'split_rules'}))
	{
		$split_rules = $TRUE;
	} else {
		$split_rules = $FALSE;
	}
}

#
# Read in rules
#
process_rules(\%usedNO, \%usedSO);
#
# Read in objects
#
if (defined($optctl{'network-objects'}) ||
	defined($optctl{'service-objects'}) ||
	defined($optctl{'properties'}))
{
	process_objects(\%objects);
}
#
# Begin print processing
#
if ($nohtml == $FALSE) {
	print "<html>\n";
	print "<title>FireWall-1 Rule Base and Object Definitions</title>\n";
	print "<body>\n";
}
#
# Print rules
#
if (defined($optctl{'rules'}))
{
	print_rules(\%usedNO, \%usedSO, $sph, $spd, $nohtml);
}
#
# Print various objects
#
if (defined($optctl{'network-objects'}) ||
	defined($optctl{'service-objects'}) ||
	defined($optctl{'properties'}))
{
	#
	# complete the list of used nw objects by exploding members
	#
	if (defined($optctl{'rules'}))
	{
		get_used_objects(\%objects, \%usedNO, \%usedSO);
	}
	#
	# Page Header
	#
	if ($nohtml == $FALSE) {
		print "<h2>FireWall-1 Object Definitions</h2>\n";
	}
}

if (defined($optctl{'network-objects'}))
{
	print_address_objs($sph, $spd, $nohtml);
	print_network_objs($sph, $spd, $nohtml);
}

if (defined($optctl{'service-objects'}))
{
	print_service_objs($sph, $spd, $nohtml);
	print_NAT_objs($sph, $spd, $nohtml);
}


if (defined($optctl{'properties'}))
{
	$key_found = $FALSE;
	$props = '';
	print_info(\%objects, 'props', 'all');
}

$nowString = localtime();

if ($nohtml == $FALSE) {
	print "<p><hr>Generated: $nowString\n";
#	print "by $SCR_NAME v$VERSION";

	# HTML footer
	print "</body>\n";
	print "</html>\n";
}
close(RULES);
close(OBJECTS);
closeStandFile();

sub tableItemHead
{
	my($nohtml) = @_;

	if ($nohtml == $FALSE) {
		print "<tr valign=top><th>";
	}
}

sub tableItemDataHead
{
	my($nohtml) = @_;

	if ($nohtml == $FALSE)
	{
		print "<tr valign=top>\n<td>";
	}
}

sub tableHead
{
	my($tableName, $nohtml) = @_;

	# Print title
	if ($nohtml == $FALSE) {
		print "<h3>$tableName</h3>\n";
		# Open table
		print '<table border=2 cellpadding=5 bgcolor="Silver">';
		tableItemHead();
	}
	else {
		openStandFile($tableName.".txt", $nohtml);
	}

}

sub tableItemHeadTail
{
	my($nohtml) = @_;
	if ($nohtml == $FALSE)
	{
		# close table head
		print "</th></tr>";
	}
	print "\n";	
}

sub tableItemDataTail
{
	my($nohtml) = @_;
	if ($nohtml == $FALSE)
	{
		# close table head
		print "</td></tr>";
	} else {
		print "\n";
	}			
}

sub tableTail
{
	my($nohtml) = @_;
	if ($nohtml == $FALSE)
	{
		print "</table>";
	}
	print "\n";
}

sub print_service_objs()
{
	my($sph, $spd, $nohtml) = @_;

	tableHead("ServiceObjects", $nohtml);
	print "Name $sph Type $sph Port $sph S_PortFrom $sph S_PortTo $sph ";
	print "Match $sph Prolog $sph Members $sph Comment";
	tableItemHeadTail($nohtml);
	
	$key_found = $FALSE;
	print_info(\%objects, 'servobj', @svoTypes);

	# Close Table
	tableTail($nohtml);
}

sub print_NAT_objs()
{
	my($sph, $spd, $nohtml) = @_;

	tableHead("NATObjects", $nohtml);
	print "Name $sph Type $sph Protocol $sph FirstPort $sph LastPort $sph Comment";
	tableItemHeadTail($nohtml);

	$key_found = $FALSE;
	print_info(\%objects, 'servobjadtr', 'all');

	tableTail($nohtml);
}

sub print_address_objs
{
	my($sph, $spd, $nohtml) = @_;

	# Print title
	tableHead("AddressObjects", $nohtml);
	print "Name $sph Type $sph FirstIPAddress $sph LastIPAddress $sph Comment";
	tableItemHeadTail($nohtml);

	$key_found = $FALSE;
	print_info(\%objects, 'netobjadtr', 'all');

	# Close Table
	tableTail($nohtml);
}

sub print_network_objs
{
	my($sph, $spd, $nohtml) = @_;

	tableHead("NetworkObjects", $nohtml);
	print "Name $sph Type $sph Location $sph FW1 $sph IPAddress $sph ";
	print "Netmask $sph Gateway $sph Members $sph Comment";
	tableItemHeadTail($nohtml);

	$key_found = $FALSE;
	print_info(\%objects, 'netobj', @nwoTypes);

	# Close Table
	tableTail($nohtml);
}

sub process_objects
{
	my ($hash) = @_;

	while (<OBJECTS>)
	{
		chop;
		s/^\s+//;

		return if (/^\)$/);

		if (/^\:([^\s]*) \(([^\)]*)$/)
		{
			# This branch recursively handles 'open' lines, i.e. lines
			# that do not end in a ')'
			# e.g. ': (CSFB-mgmt-server'

			$key = $1; $key2 = $2;

			#	  if ($key2 eq '')
			#	    {
			#	      $key2 = $key;
			#    	    }

			if ($key eq '')
			{
				$key = $key2;
			}
			if ($key2 eq '')
			{
				$key2 = $key;
			}

			#$lcKey = lc($key2);
			$lcKey = lc($key);
			$$hash{$lcKey}[0] = $key2;
			process_objects(\%{$$hash{$lcKey}[1]});
		}
		elsif (/^\:([^\s]*) (.*)$/)
		{
			# This branch recursively handles 'closed' lines, i.e. lines
			# that end in a ')'
			# e.g. ':comments ("London Management Server")'

			$key = $1; $value = $2;

			if ($key eq '')
			{
				# Group members
				$key = 'members';
			}

			$lcKey = lc($key);
			$$hash{$lcKey}[0] = $key;

			if ($$hash{$lcKey}[1] ne '')
			{
				$$hash{$lcKey}[1] .= ', ';
			}

			if ($value =~ /^\((.*)\)$/)
			{
				$value = $1;
			}

			if ($value =~ /^\"(.*)\"$/)
			{
				$value = $1;
			}
	  
			$$hash{$lcKey}[1] .= $value;
		}
	}
}

sub get_used_objects
{
	my ($objects, $usedNO, $usedSO) = @_;

	$nwObjects = $$objects{'netobj'}[1];

	foreach $name (keys(%$nwObjects))
	{
	    $NegatednwObjects = "!" . $$nwObjects{$name}[0];
	    if ((defined($$usedNO{$$nwObjects{$name}[0]})) ||
			(defined($$usedNO{$NegatednwObjects})))
	    {
			if (defined($$nwObjects{$name}[1]{'members'}))
			{
				@members = split(',', $$nwObjects{$name}[1]{'members'}[1]);

				foreach $member (@members)
				{
				    $member =~ s/^\s*(.*)/$1/;
				    $usedNO{$member} = 1;
				}
			}
		}
	}

	$svObjects = $$objects{'servobj'}[1];

	foreach $name (keys(%$svObjects))
	{
	    $NegatedsvObjects = "!" . $$svObjects{$name}[0];
	    if ((defined($$usedSO{$$svObjects{$name}[0]})) ||
			(defined($$usedSO{$NegatedsvObjects})))
		{
			if (defined($$svObjects{$name}[1]{'members'}))
			{
				@members = split(',', $$svObjects{$name}[1]{'members'}[1]);
				foreach $member (@members)
				{
					$member =~ s/^\s*(.*)/$1/;
					$usedSO{$member} = 1;
				}
			}
		}
	}
}


sub print_info(\%$@)
{
	my ($hash, $key, @types) = @_;
	my ($k, $actType);

	if ($key_found == $TRUE && $key eq 'props')
	{
		print_properties($hash, $key);
	}
	else
	{
		foreach $k (sort keys(%$hash))
		{
			if ($key_found == $TRUE)
			{
				if ($types[0] ne 'all')
				{
					$actType = lc($$hash{$k}[1]{type}[1]);
					next if (!grep(/^$actType$/, @types));
				}

				$NegateHash="!"."$$hash{$k}[0]";

				if ($key eq 'netobj')
				{
					# this $all_objs is made here to list ALL objs instead of 'in use'. wxu
				    if (!$rules || defined($usedNO{$$hash{$k}[0]})
						|| defined($usedNO{$NegateHash}) || $all_objs)
					{
						print_network_object($hash, $k);
					}
				}
				elsif ($key eq 'servobj')
				{
					if (!$rules || defined($usedSO{$$hash{$k}[0]})
						|| defined($usedSO{$NegateHash}) || $all_objs)
					{
						print_service_object($hash, $k);
					}
				}
				elsif ($key eq 'netobjadtr')
				{
					if (!$rules || defined($usedNO{$$hash{$k}[0]})
						|| defined($usedNO{$NegateHash}))
					{
						print_adtr_network_object($hash, $k);
					}
				}		  
				elsif ($key eq 'servobjadtr')
				{
					if (!$rules || defined($usedSO{$$hash{$k}[0]})
						|| defined($usedSO{$NegateHash}))
					{
						print_adtr_service_object($hash, $k);
					}
				}
			}
			else
			{
				if ($$hash{$k}[0] eq $key)
				{
					$key_found = $TRUE;
				}
				if (ref($$hash{$k}[1]))
				{	
					print_info($$hash{$k}[1], $key, @types);
				}
			}
		}
	}
}


sub print_network_object
{
	my($hash, $key) = @_;

	$pName = $pLocation = $pIpAddress = $pNetmask = $pGateway = '-';
	$pMembers = $pComment = $pFwInstalled = '-';
	$sp = ",";

	# Name
	# FW-1 colors are affixed to the object names.
	# Html anchors added so group members can refer back to the objects.
	$pName = $$hash{$key}[0];
	if ($nohtml == FALSE)
	{
		$sp = "<br>";
		$pNameLink = '<a NAME="' . $pName . '"></a>';
		if ( $with_colors == $TRUE )
		{
    		 $pNameLink .= $colors{lc($$hash{$key}[1]{color}[1])};
		}
		$pName = $pNameLink . $pName;
	}

	# Type
	$pType = lc($$hash{$key}[1]{type}[1]);

	return if (!grep(/^$pType$/, @nwoTypes));

	if ($pType eq 'host')
	{
		$pType = 'Host';
		$pFwInstalled = 'No';
	}
	elsif ($pType eq 'gateway')
	{
		$pType = 'Gateway';
	}
	elsif ($pType eq 'router')
	{
		$pType = 'Router';
		$pFwInstalled = 'No';
	}
	elsif ($pType eq 'switch')
	{
		$pType = 'Switch';
		$pFwInstalled = 'No';
	}
	elsif ($pType eq 'network')
	{
		$pType = 'Network';
	}
	elsif ($pType eq 'domain')
	{
		$pType = 'Domain';
	}
	elsif ($pType eq 'logical')
	{
		$pType = 'Logical';
	}
	elsif ($pType eq 'group')
	{
		$pType = 'Group';
	}
	elsif ($pType eq 'blackbox')
	{
		$pType = 'Blackbox';
	}

	# Location
	if (defined($$hash{$key}[1]{location}))
	{
		if (lc($$hash{$key}[1]{location}[1]) eq 'external')
		{
			$pLocation = 'External';
		}
		elsif (lc($$hash{$key}[1]{location}[1]) eq 'internal')
		{
			$pLocation = 'Internal';
		}
	}

	# Firewall installed?
	if (defined($$hash{$key}[1]{firewall}))
	{
		if (lc($$hash{$key}[1]{firewall}[1]) eq 'installed')
		{
			$pFwInstalled = 'Yes';
		}
	}

	# IP-Address
	if (defined($$hash{$key}[1]{ipaddr}))
	{
		$pIpAddress = checkEmptyString($$hash{$key}[1]{ipaddr}[1]);
	}

	# Netmask
	if (defined($$hash{$key}[1]{netmask}))
	{
		$pNetmask = checkEmptyString($$hash{$key}[1]{netmask}[1]);
	}

	# Gateway
	if (defined($$hash{$key}[1]{gateways}) &&
	$$hash{$key}[1]{gateways}[1] !~ /^\s*$/)
	{
		$pIpGateway = checkEmptyString($$hash{$key}[1]{gateways}[1]);
	}

	# Members
	if (defined($$hash{$key}[1]{members}))
	{
		$pMembers = checkEmptyString($$hash{$key}[1]{members}[1]);
		# Remove white space
		$pMembers =~ s/\s*//g;
		if ($nohtml == $FALSE)
		{
			# Provide links to object definitions
			$pMembers =~ s/([^,]+)/\<a href\=\"\#\1\"\>\1\<\/a\>/g;
			# Replace ',' with <br>
			$pMembers =~ s/,\s*/<br>/g;
		}
	}

	# Comments
	if (defined($$hash{$key}[1]{comments}))
	{
		$pComments = checkEmptyString($$hash{$key}[1]{comments}[1]);
	}

	# Interfaces?
	foreach $k (sort keys(%{$$hash{$key}[1]}))
	{
		if (ref($$hash{$key}[1]{$k}[1]) &&
		defined($$hash{$key}[1]{$k}[1]{iffullname}))
		{
			$pName .= '$sp'.$$hash{$key}[1]{$k}[1]{iffullname}[1];

			if (defined($$hash{$key}[1]{$k}[1]{ipaddr}))
			{
				$pIpAddress .= '$sp'.$$hash{$key}[1]{$k}[1]{ipaddr}[1];
			}

			if (defined($$hash{$key}[1]{$k}[1]{netmask}))
			{
				$pNetmask .= '$sp'.$$hash{$key}[1]{$k}[1]{netmask}[1];
			}
		}
	}

	tableItemDataHead($nohtml);
	print "$pName $spd $pType $spd $pLocation $spd $pFwInstalled $spd";
	print "$pIpAddress $spd $pNetmask $spd $pGateway $spd $pMembers $spd";
	print "$pComments";
	tableItemDataTail($nohtml);
}

sub print_adtr_network_object
{
	my($hash, $key) = @_;

	$pComment = '-';

	# Name
	$pName = $$hash{$key}[0];

	# Type
	$pType = lc($$hash{$key}[1]{type}[1]);

	return if (!grep(/^$pType$/, @nwoTypes));

	if ($pType eq 'machines_range')
	{
		$pType = 'Range';
	}

	# IP-Address
	if (defined($$hash{$key}[1]{ipaddr_first}))
	{
		$pIpFirstAddress = checkEmptyString($$hash{$key}[1]{ipaddr_first}[1]);
	}

	# IP-Address
	if (defined($$hash{$key}[1]{ipaddr_last}))
	{
		$pIpLastAddress = checkEmptyString($$hash{$key}[1]{ipaddr_last}[1]);
	}

	# Comments
	if (defined($$hash{$key}[1]{comments}))
	{
		$pComments = checkEmptyString($$hash{$key}[1]{comments}[1]);
	}

	tableItemDataHead($nohtml);
	print "$pName $spd $pType $spd $pIpFirstAddress $spd $pIpLastAddress $spd";
	print "$pComments";
	tableItemDataTail($nohtml);

}

sub print_service_object
{
	my ($hash, $key) = @_;

	$pPort = $pFrom = $pTo = $pMatch = $pProlog = $pMembers = $pComments = '-';

	# Name
	# FW-1 colors are affixed to the object names.
	# Html anchors added so group members can refer back to the objects.

	$pName = $$hash{$key}[0];
	if ($nohtml == FALSE)
	{
		$pNameLink = '<a NAME="' . $pName . '"></a>';
		if ( $with_colors == $TRUE )
		{
    		 $pNameLink .= $colors{lc($$hash{$key}[1]{color}[1])};
		}
		$pName = $pNameLink . $pName;
	}

	# Type
	$pType = lc($$hash{$key}[1]{type}[1]);

	return if (!grep(/^$pType$/, @svoTypes));

	# Port
	if (defined($$hash{$key}[1]{port}))
	{
		$pPort = checkEmptyString($$hash{$key}[1]{port}[1]);
	}

	# Source port range / from
	if (defined($$hash{$key}[1]{src_port_from}))
	{
		$pFrom = checkEmptyString($$hash{$key}[1]{src_port_from}[1]);
	}

	# Source port range / to
	if (defined($$hash{$key}[1]{src_port_to}))
	{
		$pTo = checkEmptyString($$hash{$key}[1]{src_port_to}[1]);
	}

	# Match expression
	if (defined($$hash{$key}[1]{'exp'}))
	{
		$pMatch = checkEmptyString($$hash{$key}[1]{'exp'}[1]);
	}

	# Prolog
	if (defined($$hash{$key}[1]{'prolog'}))
	{
		$pProlog = checkEmptyString($$hash{$key}[1]{'prolog'}[1]);
	}

	# Group members
	if (defined($$hash{$key}[1]{members}))
	{
		$pMembers = $$hash{$key}[1]{members}[1];
		# Remove white space
		$pMembers =~ s/\s*//g;
		# Provide links to object definitions
		if ( $nohtml == $FALSE)
		{
			$pMembers =~ s/([^,]+)/\<a href\=\"\#\1\"\>\1\<\/a\>/g;
			# Replace ',' with <br>
			$pMembers =~ s/,\s*/<br>/g;
		}
		
		$pMembers = checkEmptyString($pMembers);
	}

	# Comments
	if (defined($$hash{$key}[1]{'comments'}))
	{
		$pComments = checkEmptyString($$hash{$key}[1]{'comments'}[1]);
	}

	tableItemDataHead($nohtml);
	print "$pName $spd $pType $spd $pPort $spd $pFrom $spd $pTo $spd";
	print "$pMatch $spd $pProlog $spd $pMembers $spd $pComments";
	tableItemDataTail($nohtml);
}

sub print_adtr_service_object
{
	my ($hash, $key) = @_;

	$pComment = '-';

	# Name
	$pName = $$hash{$key}[0];

	# Type
	$pType = lc($$hash{$key}[1]{type}[1]);

	return if (!grep(/^$pType$/, @svoTypes));

	if ($pType == 'port_Range')
	{
		$pType = 'Range';
	}

	# Protocol
	if (defined($$hash{$key}[1]{ports_range_type}))
	{
		$pProtocol = checkEmptyString($$hash{$key}[1]{ports_range_type}[1]);
	}

	# First Port
	if (defined($$hash{$key}[1]{port_first}))
	{
		$pFirstPort = checkEmptyString($$hash{$key}[1]{port_first}[1]);
	}

	# Last Port
	if (defined($$hash{$key}[1]{port_last}))
	{
		$pLastPort = checkEmptyString($$hash{$key}[1]{port_last}[1]);
	}

	# Comments
	if (defined($$hash{$key}[1]{'comments'}))
	{
		$pComments = checkEmptyString($$hash{$key}[1]{'comments'}[1]);
	}

	tableItemDataHead($nohtml);
	print "$pName $spd $pType $spd $pProtocol $spd $pFirstPort $spd";
	print "$pLastPort $spd $pComments";
	tableItemDataTail($nohtml);
}

sub get_logtrack_value
{
	my ($hash, $propName) = @_;

	if (defined($$hash{$propName}[1]{type}))
	{
		return($$hash{$propName}[1]{type}[1]);
	}
	else
	{
		return('none');
	}
}

sub print_properties
{
	my($hash, $key) = @_;

	if ($nohtml == $FALSE) {
		$sp = "<br>";
	} else {
		$sp = ",";
	}

	if (defined($$hash{auth}))
	{
		$pAuthFailureTrack = 'auth';
	}
	elsif (defined($$hash{authalert}))
	{
		$pAuthFailureTrack = 'alert';
	}
	else
	{
		$pAuthFailureTrack = 'none';
	}

	# Print title
	if ($nohtml == $FALSE) {
		print "<h3>Properties</h3>\n";
	}

	# Open table
	tableHead("PropertiesSecurityPolicy", $nohtml);
	print "Property $sph Setting $sph Values";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "Apply Gateway Rules to Interface Direction $spd n/a $spd";
	print checkEmptyString($$hash{gatewaydir}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "TCP Session Timeout (sec) $spd n/a $spd";
	print checkEmptyString($$hash{tcptimeout}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Firewall-1 Control Connections $spd";
	print checkEmptyString($$hash{fw1enable}[1])."$spd";
	print checkEmptyString($$hash{fw1_enable_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept UDP Replies $spd";
	print checkEmptyString($$hash{udpreply}[1])."$spd";
	print checkEmptyString($$hash{udpreply_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "UDP Reply Timeout (sec) $spd n/a $spd ";
	print checkEmptyString($$hash{udptimeout}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Outgoing Packets";
	print checkEmptyString($$hash{outgoing}[1])."$spd";
	print checkEmptyString($$hash{outgoing_p}[1])."$spd n/a";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable Decryption on Accept $spd";
	print checkEmptyString($$hash{acceptdecrypt}[1])."$spd n/a";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Use FASTPATH $spd";
	print checkEmptyString($$hash{enable_fastpath}[1])."$spd n/a";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept RIP $spd";
	print checkEmptyString($$hash{rip}[1])."$spd";
	print checkEmptyString($$hash{rip_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Domain Name Queries (UDP) $spd";
	print checkEmptyString($$hash{domain_udp}[1])."$spd";
	print checkEmptyString($$hash{domain_udp_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Domain Name Download (TCP) $spd";
	print checkEmptyString($$hash{domain_tcp}[1])."$spd";
	print checkEmptyString($$hash{domain_tcp_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept ICMP $spd";
	print checkEmptyString($$hash{icmpenable}[1])."$spd";
	print checkEmptyString($$hash{icmpenable_p}[1]);
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesServices", $nohtml);
	print "Property $sph Setting $sph Values";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable FTP PASV Connections $spd";
	print checkEmptyString($$hash{ftppasv}[1])."$spd";
	print checkEmptyString($$hash{ftppasv_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable RSH/REXEC Reverse stderr Connections $spd";
	print checkEmptyString($$hash{rshstderr}[1])."$spd";
	print checkEmptyString($$hash{rshstderr_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable RPC Control $spd";
	print checkEmptyString($$hash{rpcenable}[1])."$spd";
	print checkEmptyString($$hash{rpcenable_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable Response of FTP Data Connections $spd";
	print checkEmptyString($$hash{ftpdata}[1])."$spd";
	print checkEmptyString($$hash{ftpdata_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable Real Audio Reverse Connections";
	print checkEmptyString($$hash{raudioenable}[1])."$spd";
	print checkEmptyString($$hash{raudioenable_p}[1])."$spd n/a";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable VDOLive Reverse Connections $spd";
	print checkEmptyString($$hash{vdolivenable}[1])."$spd";
	print checkEmptyString($$hash{vdolivenable_p}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable CoolTalk Data Connections (UDP) $spd";
	print checkEmptyString($$hash{cooltalkenable}[1])."$spd n/a";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable H.323 Control and Data Connections $spd";
	print checkEmptyString($$hash{iphoneenable}[1])."$spd n/a";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesLogAlert", $nohtml);
	print "Property $sph Setting";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "Excessive Log Grace Period (sec) $spd $$hash{loggrace}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "PopUp Alert Command $spd $$hash{alertcmd}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Mail Alert Command $spd $$hash{mailcmd}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "SNMP Trap Alert Command $spd $$hash{snmptrapcmd}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "User Defined Alert Command $spd $$hash{useralertcmd}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Anti Spoof Alert Command $spd $$hash{spoofalertcmd}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "User Authentication Alert Command $spd $$hash{userauthalertcmd}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	$ipOptsLog = get_logtrack_value($hash, 'ipoptslog');
	print "IP Options Drop Track $spd $ipOptsLog";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Log Established TCP Packets $spd $$hash{log_established_tcp}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Enable Active Connections $spd $$hash{liveconns}[1]";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesResolving", $nohtml);
	print "Property $sph Setting";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "Lookup Priorities $spd";
	print "1. $$hash{resolver_1}[1]"."$sp";
	print "2. $$hash{resolver_2}[1]"."$sp";
	print "3. $$hash{resolver_3}[1]"."$sp";
	print "4. $$hash{resolver_4}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "BIND Timeout (sec) $spd $$hash{timeout}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "BIND Retries $spd $$hash{retries}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Log Viewer Resolver Properties $spd";
	print "Page Timeout (sec): $$hash{pagetimeout}[1]";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesSecurityServers", $nohtml);
	print "Property $sph Setting";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "Telnet Welcome Message File $spd";
	print checkEmptyString($$hash{telnet_msg}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Rlogin Welcome Message File $spd";
	print checkEmptyString($$hash{rlogin_msg}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "FTP Welcome Message File $spd";
	print checkEmptyString($$hash{ftp_msg}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Client Authentication Welcome Message File $spd";
	print checkEmptyString($$hash{clnt_auth_msg}[1]);
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "HTTP Next Proxy $spd";
	print "Host: $$hash{http_next_proxy_host}[1]"."$sp ";
	print "Port: $$hash{http_next_proxy_port}[1]";
	tableItemDataTail($nohtml);

	if (defined($$hash{http_servers}))
	{
		$servers = $$hash{http_servers}[1];

		tableItemDataHead($nohtml);
		print "<HTTP Servers $spd";

		foreach $srv (keys(%$servers))
		{
			print "Name: $$servers{$srv}[0]"."$sp";
			print "Host: $$servers{$srv}[1]{host}[1]"."$sp";
			print "Port: $$servers{$srv}[1]{port}[1]"."$sp";
			print "Null Server: $$servers{$srv}[1]{ers}[1]"."$sp";
			print "Reauthentication: $$servers{$srv}[1]{reauth_requests}[1]";
		}

		tableItemDataTail($nohtml);
	}

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesAuthentications", $nohtml);
	print "Property $sph Setting";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "User Authentication Session Timeout (min)"."$spd";
	print "$$hash{au_timeout}[1]";
	tableItemDataTail($nohtml);

	$pAuthFailureTrack = get_logtrack_value($hash, default_track);

	tableItemDataHead($nohtml);
	print "Authentication Failure Track"."$spd";
	print "$pAuthFailureTrack";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "AXENT Pathways Defender Server Setup"."$spd";
	print "Server IP: $$hash{snk_server_ip}[1]"."$sp ";
	print "Agent ID: $$hash{snk_agent_id}[1]"."$sp ";
	print "Agent Key: $$hash{snk_agent_key}[1]";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesSYNDefender", $nohtml);
	print "Property $sph Setting";
	tableItemHeadTail($nohtml);

	if ($$hash{fwsynatk_method}[1] == 0)
	{
		$pFwSynAtkMethod = 'None';
	}
	elsif ($$hash{fwsynatk_method}[1] == 1)
	{
		$pFwSynAtkMethod = 'SYN Relay';
	}
	elsif ($$hash{fwsynatk_method}[1] == 2)
	{
		$pFwSynAtkMethod = 'SYN Gateway';
	}
	elsif ($$hash{fwsynatk_method}[1] == 3)
	{
		$pFwSynAtkMethod = 'Passive SYN Gateway';
	}

	tableItemDataHead($nohtml);
	print "Method $spd $pFwSynAtkMethod";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Timeout"."$spd $$hash{fwsynatk_timeout}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Maximum Sessions"."$spd $$hash{fwsynatk_max}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Display Warning Messages"."$spd";
	print "$$hash{fwsynatk_warning}[1]";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesMiscellaneous", $nohtml);
	print "Property $sph Setting";
	tableItemHeadTail($nohtml);

	tableItemDataHead($nohtml);
	print "Load Agents Port $spd $$hash{load_service_port}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Load Measurement Interval $spd";
	print "$$hash{lbalanced_period_wakeup_sec}[1]";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);

	# Open table
	tableHead("PropertiesAccessLists", $nohtml);
	print "Property $sph Setting $sph Values";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Established TCP Connections $spd";
	print "$$hash{established_router}[1] $spd $$hash{established_router_p}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept RIP $spd";
	print "$$hash{rip_router}[1] $spd $$hash{rip_router_p}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Domain Name Queries (UDP) $spd";
	print "$$hash{domain_udp_router}[1] $spd $$hash{domain_udp_router_p}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept Domain Name Download (TCP) $spd";
	print "$$hash{domain_tcp_router}[1] $spd $$hash{domain_tcp_router_p}[1]";
	tableItemDataTail($nohtml);

	tableItemDataHead($nohtml);
	print "Accept ICMP $spd";
	print "$$hash{icmpenable_router}[1] $spd $$hash{icmpenable_router_p}[1]";
	tableItemDataTail($nohtml);

	# Close table
	tableTail($nohtml);
}

sub checkEmptyString()
{
	my ($inputSt) = @_;
	$ret = "-";
	if ( !($inputSt eq "") )
	{
#		if ($nohtml == $FALSE) {
#			$ret = "&nbsp";
#		}
#	} else {
		$ret = $inputSt;
	} 
	return $ret;
}

sub process_rules
{
	my ($objects) = @_[0];
	my ($services) = @_[1];

	$mode = -1;
	$ruleMode = 0;
	$adtrRuleMode = 1;

	$ruleNr = 0;
	$adtrRuleNr = 0;

	$printFlag = 0;

	while (<RULES>)
	{
		$line++;
		s/^\s+//;

		if (/^\(\"\#\#(.+)\"/)
		{
			$policyName = $1;
		}
		elsif (/^\(/ || /^\)/)
		{
			next;
		}
		elsif (/^\:rule /)
		{
			$ruleNr++;
			$ruleTable[$ruleNr][0][0] = $ruleNr;
			$mode = $ruleMode;
		}
		elsif (/^\:rule_adtr /)
		{
			$adtrRuleNr++;
			$adtrRuleTable[$adtrRuleNr][0][0] = $adtrRuleNr;
			$mode = $adtrRuleMode;
		}
		elsif ($mode == $ruleMode)
		{
			if (/^\:op \(\"not in\"\)/)
			{
				$not = '!';

				if ($entryNr > 0)
				{
					for ($i = 0; $i < $entryNr; $i++)
					{
						$ruleTable[$ruleNr][$column][$i] =
						$not . $ruleTable[$ruleNr][$column][$i];
					}
				}
			}
			elsif (/^\:(\w+) /)
			{
				$col = $1;
	
				if (defined($fieldIndices{$col}))
				{
					$column = $fieldIndices{$col};
				}
				else
				{
					next;
				}

				$entryNr = 0;
				$not = '';

				if ($col eq 'disabled')
				{
					/^\:disabled\s*\((.*)\)\n$/;
					$value = $1;

					$ruleTable[$ruleNr][$column][$entryNr++] = $value;
				}

				if ($col eq 'comments')
				{
					/^\:comments\s*(.*)\n$/;
					$value = $1;

					if ($value ne '()')
					{
						$value =~ s/^\(\"(.*)\"\)/$1/;
						if ($nohtml == $FALSE) {
							$value =~ s/;/<br>/g;
						} else {
							$value =~ s/;/ /g;
						}
						$ruleTable[$ruleNr][$column][$entryNr++] = $value;
					}
				}
			}
			elsif (/^\: ([^)]+)$/)
			{
				chomp($value = $1);

				if ($value =~ /^\(/)
				{
					$value =~ s/^\(\"?([^\"]+)\"?$/$1/;
					do
					{
						$_ = <RULES>;
					}
					#
					# This fixes the problem with Session Authentication rules.
					# I don't like this because the script is now more dependant
					# on the syntax of the rules file. SON - Oct 1999
					# 
					while ( ! /^\t\t\)$/ );
					#while (/^\:/);
				}

				$ruleTable[$ruleNr][$column][$entryNr++] = $not . $value;
			}
		}
		elsif ($mode == $adtrRuleMode)
		{
			if (/^\:(\w+) /)
			{
				$col = $1;

				if ($col eq 'adtr_method')
				{
					if ($adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] ne 'Any')
					{
						if (/static/)
						{
							$adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] =
							'S:'.$adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1];
						}
						elsif (/hide/)
						{
							$adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] =
							'H:'. $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1];
						}
						else
						{
							$adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] =
							'U:'. $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1];
						}
					}

					next;
				}
	     
				if (defined($adtrFieldIndices{$col}))
				{
					$column = $adtrFieldIndices{$col};
				}
				else
				{
					next;
				}

				$entryNr = 0;

				if ($col eq 'comments')
				{
					/^\:comments\s*(.*)\n$/;
					$value = $1;

					if ($value ne '()')
					{
						$value =~ s/^\(\"(.*)\"\)/$1/;
						if ($nohtml == $FALSE) {
							$value =~ s/;/<br>/g;
						} else {
							$value =~ s/;/ /g;
						}
						$adtrRuleTable[$adtrRuleNr][$column][$entryNr++] = $value;
					}
				}
			}
			elsif (/^\: (.+)/)
			{
				$value = $1;

				if ($value =~ /^\(/)
				{
					$value =~ s/^\(\"?([^\"]+)\"?$/$1/;
					do
					{
						$_ = <RULES>;
					}
					while (/^\:/);
				}

				$adtrRuleTable[$adtrRuleNr][$column][$entryNr++] = $value;
			}
		}
	}
}

sub print_icon
{
	my ($objects, $object_textname) = @_;

	#
	# Figure out icon to place in front of object text name
	#
	# Check on non-user defined objects first, basically reserved word
	# things like 'Any', 'Gateways', etc.
	#
	$negated = 0;
	#
	if ( $object_textname =~ /^\!(.*)/ )
	{
		$object_textname = $1;
		$negated = 1;
	}
	$textObject_lc = lc( $object_textname );
	if ( $without_icons == $FALSE )
	{
		#
		if ( $textObject_lc eq 'any')
		{
		 	print "<image src=\"$ICONPATH/any.gif\">";
		}
		elsif ( $textObject_lc eq 'gateways')
		{
		 	print "<image src=\"$ICONPATH/gateways.gif\">";
		}
		elsif ( $textObject_lc eq 'drop')
		{
		 	print "<image src=\"$ICONPATH/drop.gif\">";
		}
		elsif ( $textObject_lc eq 'short')
		{
		 	print "<image src=\"$ICONPATH/short.gif\">";
		}
		elsif ( $textObject_lc eq 'long')
		{
		 	print "<image src=\"$ICONPATH/long.gif\">";
		}
		elsif ( $textObject_lc eq 'accept')
		{
		 	print "<image src=\"$ICONPATH/accept.gif\">";
		}
		elsif ( $textObject_lc eq 'reject')
		{
		 	print "<image src=\"$ICONPATH/reject.gif\">";
		}
		elsif ( $textObject_lc eq 'client encrypt')
		{
		 	print "<image src=\"$ICONPATH/encrypt.gif\">";
		}
		elsif ( $textObject_lc eq 'client auth')
		{
		 	print "<image src=\"$ICONPATH/clientauth.gif\">";
		}
		elsif ( $textObject_lc eq 'session auth')
		{
		 	print "<image src=\"$ICONPATH/sessionauth.gif\">";
		}
		elsif ( $textObject_lc eq 'account')
		{
		 	print "<image src=\"$ICONPATH/account.gif\">";
		}
		elsif ( $textObject_lc eq 'encrypt')
		{
		 	print "<image src=\"$ICONPATH/encrypt.gif\">";
		}
		#
		# Check on non-user defined objects first, basically reserved word
		# things like 'Any', 'Gateways', etc.
		#
		#
		# Host object
		#
		elsif ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'host' )
		{
			if ( $negated == 1 )
			{
				print "<a border=0 href=\"#${object_textname}\"><image src=\"$ICONPATH/negatehost.gif\"></a>";
			} else {
				print "<a border=0 href=\"#${object_textname}\"><image src=\"$ICONPATH/host.gif\"></a>";
			}
		}
		#
		# Firewall object
		#
		elsif ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'gateway' )
		{
			if ( $negated == 1 )
			{
				if ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'location'}[1]) eq 'internal' )
				{
					print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/negatefirewallinternal.gif\"></a>";
				} else {
					print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/negatefirewallremote.gif\"></a>";
				}
			} else {
				if ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'location'}[1]) eq 'internal' )
				{
					print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/firewallinternal.gif\"></a>";
				} else {
					print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/firewallremote.gif\"></a>";
				}
			}
		}
		#
		# Group object
		#
		elsif ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'group' )
		{
			if ( $negated == 1 )
			{
				print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/negategroup.gif\"></a>";
			} else {
				print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/group.gif\"></a>";
			}
		}
		#
		# Network object
		#
		elsif ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'network' )
		{
			if ( $negated == 1 )
			{
				print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/negatenetwork.gif\"></a>";
			} else {
				print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/network.gif\"></a>";
			}
		}
		#
		# Router object
		#
		elsif ( lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'router' )
		{
			if ( $negated == 1 )
			{
				print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/negaterouter.gif\"></a>";
			} else {
				print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/router.gif\"></a>";
			}
		}
		#
		# User group object
		#
		elsif ( $textObject_lc =~ /\@/ )
		{
			#print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/usergroup.gif\"></a>";
			print "<image src=\"$ICONPATH/usergroup.gif\">";
		}
		#
		# Group service object
		#
		elsif ( lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'group' )
		{
			print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/group.gif\"></a>";
		}
		#
		# TCP service object
		#
		elsif ( lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'tcp' )
		{
			print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/tcp.gif\"></a>";
		}
		#
		# UDP service object
		#
		elsif ( lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'udp' )
		{
			print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/udp.gif\"></a>";
		}
		#
		# RPC service object
		#
		elsif ( lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'rpc' )
		{
			print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/rpc.gif\"></a>";
		}
		#
		# ICMP service object
		#
		elsif ( lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'icmp' )
		{
			print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/icmp.gif\"></a>";
		}
		#
		# Other service object
		#
		elsif ( lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]) eq 'other' )
		{
			print "<a href=\"#${object_textname}\"><image src=\"$ICONPATH/otherservice.gif\"></a>";
		}
	}
	return $object_textname;
}

sub print_entry
{
	my ($objects, $line, $row, $entryNr) = @_;

		$object_textname = $ruleTable[$line][$row][$entryNr];
		$object_textname = print_icon($objects, $object_textname);
		#
		# Print object
		#
		print $object_textname;
		if ( $row == 0 && defined($ruleTable[$line][$fieldIndices{'disabled'}]) ) {
			if ($nohtml == $FALSE) {
				print "<BR><BR>";
			}
			print " Disabled";
		}
}

sub print_cell_entry
{
	my ($objects, $line, $row, $maxRow, $entryNr) = @_;

	print_entry($objects, $line, $row, $entryNr);

	if ($row < $maxRows) {
		print "$spd";
	}
}


sub print_row_entries
{
	my ($objects, $services, $line, $row, $maxRow, $reset) = @_;

		$maxNr = $#{$ruleTable[$line][$row]} + 1;

		if ($maxNr >= 1) {
			for ($entryNr = 0; $entryNr < $maxNr; $entryNr++)
			{
				if ($entryNr > 0)
				{
					if ($nohtml == $FALSE) {
						print "<br>";
					} else {
						print ",";
					}
				}
				print_entry($objects, $line, $row, $entryNr);
				if ($rest == $TRUE)
				{
					if (($row == 1) || ($row == 2))
					{
						$$objects{$ruleTable[$line][$row][$entryNr]} = 1;
					}
					elsif ($row == 3)
					{
						$$services{$ruleTable[$line][$row][$entryNr]} = 1;
					}
				}
			}
		} else 
		{
			print "-";
		}

		if ($row < $maxRows) {
			print "$spd";
		}
}

sub print_line_3split_entry
{
	my ($objects, $services, $line, $maxRows) = @_;

	$maxSrc = $#{$ruleTable[$line][$fieldIndices{'src'}]} + 1;
	$maxDst = $#{$ruleTable[$line][$fieldIndices{'dst'}]} + 1;
	$maxSvr = $#{$ruleTable[$line][$fieldIndices{'services'}]} + 1;

	for ($iSrc = 0; $iSrc < $maxSrc; $iSrc++) 
	{
		for ($iDst = 0; $iDst < $maxDst; $iDst++) 
		{
			for ($iSvr = 0; $iSvr < $maxSvr; $iSvr++) 
			{
				for ($row = 0; $row <= $maxRows; $row++)
				{
					if ($row == $fieldIndices{'src'})
					{
						print_cell_entry($objects, $line, $row, $maxRows, $iSrc);
					}
					elsif ($row == $fieldIndices{'dst'})
					{
						print_cell_entry($objects, $line, $row, $maxRows, $iDst);
					}
					elsif ($row == $fieldIndices{'services'})
					{
						print_cell_entry($objects, $line, $row, $maxRows, $iSvr);
					} else 
					{
						print_row_entries($objects, $services, $line, $row, $maxRows, $FALSE);
					}
				}
				tableItemDataTail($nohtml);
			}
		}
	}
}


sub print_rules
{
	my ($objects, $services, $sph, $spd, $nohtml) = @_;

	my ($textObject_lc);
	my ($object_textname);
	my ($negated) = 0;
	my ($printFlag) = 0;

	# Page Header
	tableHead("FirewallPolicy", $nohtml);

	$maxRows = $#headers;

	# Table headers
	for ($row = 0; $row <= $maxRows; $row++)
	{
		print "$headers[$row]";
		if ( $row < $maxRows ) {
			print "$sph";
		}
	}
	tableItemHeadTail($nohtml);

	for ($line = 1; $line <= $ruleNr; $line++)
	{
		if (!defined($ruleTable[$line][$fieldIndices{'time'}][0]))
		{
			$ruleTable[$line][$fieldIndices{'time'}][0] = 'Any';
		}
      
		if ($nohtml == $FALSE) {
			print "<tr valign=top><td>";
		}
		if ($split_rules == $TRUE)
		{
			print_line_3split_entry($objects, $services, $line, $maxRows);
		} else {
			for ($row = 0; $row <= $maxRows; $row++)
			{
				print_row_entries($objects, $services, $line, $row, $maxRows, $TRUE);
			}
			tableItemDataTail($nohtml);
		}
	}

	# Close table
	tableTail($nohtml);

	if ($adtrRuleNr)
	{
		# Title
		tableHead("AddressTranslationRules", $nohtml);

		# Table headers
		$maxRows = $#adtrHeaders;
		for ($row = 0; $row <= $maxRows; $row++)
		{
			print "$adtrHeaders[$row]";
			if ( $row < $maxRows ) {
				print "$sph";
			}
		}
		tableItemHeadTail($nohtml);

		#
		# Reset $printFlag to zero
		#
		$printFlag = 0;
		#
		for ($line = 1; $line <= $adtrRuleNr; $line++)
		{
			if ( $nohtml == $FALSE ) {
				print "<tr valign=top><td>";
			}
			for ($row = 0; $row <= $maxRows; $row++)
			{
				$x = $#{$adtrRuleTable[$line][$row]} + 1;

				for ($entryNr = 0; $entryNr < $x; $entryNr++)
				{
					if ($entryNr > 0)
					{
						if ( $nohtml == $FALSE ) {
							print "<br>";
						} else {
							print ",";
						}
					}

					$printFlag = 1;

					print "$adtrRuleTable[$line][$row][$entryNr]";

					if (($row == 1) || ($row == 2) || ($row == 4) || ($row == 5))
					{
						if ($adtrRuleTable[$line][$row][$entryNr] =~ /.?\:(.+)/)
						{
							$adtrRuleTable[$line][$row][$entryNr] = $1;
						}

						$$objects{$adtrRuleTable[$line][$row][$entryNr]} = 1;
					}
					elsif (($row == 3) || ($row == 6))
					{
						$$services{$adtrRuleTable[$line][$row][$entryNr]} = 1;
					}
				}

				if ($printFlag == 0 && $nohtml == $FALSE)
				{
					print "&nbsp"
				}
				if ($row < $maxRows) {
					print "$spd";
				}
				$printFlag = 0;
			}

			tableItemDataTail($nohtml);
		}

		# Close table
		tableTail($nohtml);
	}
}

sub closeStandFile
{
	close(STDOUT);
}

sub openStandFile
{
	my($fileName, $nohtml) = @_;

	closeStandFile();
	if (!open(STDOUT, ">".$fileName))
	{
		print STDERR "$SCR_NAME: Couldn't open output file $fileName\n";
		exit 1;
	}
}

sub usage
{
  print STDERR
  "usage: $SCR_NAME { --all | --network-objects [--tn=<type-list>] |
		   --service-objects [--ts=<type-list>] | properties}
                   [--objects=<objects file>] [--rules=<rule file>]
				   [--color] [<output file>]
       $SCR_NAME --help
       $SCR_NAME --version

   --all: List all objects (network objects, service objects and properties).
          This is the default.

   --network-objects: Include network objects in the output. See also '--tn'.
          If a rule base is specified, only network objects that appear in
          that rule base will be included in the output.

   --objects=<objects file>: Location of FireWall-1 objects file. Default is
          'objects.C'.

   --properties: Include property settings in the output.

   --service-objects: Include service objects in the output. See also '--ts'.
          If a rule base is specified, only service objects that appear in
          that rule base will be included in the output.

   --tn=<type-list>: List of network object types that are included in the
          output. See also 'type-list' below.

   --ts=<type-list>: List of service object types that are included in the
          output. See also 'type-list' below.

   --rules=<rule file>: Location of FireWall-1 rule file. If specified,
          FireWall-1 rule base will be included in output.

   --with_color: Add html color tags to network and service objects to reflect
          the colors used in the FW-1 GUI.  By default, this is OFF.

   --split_rules: multiple rules under one rule number will be splited. 
			By default, this is OFF.

   --all_objs: it will list ALL services and/or objects. By defailt only those
			IN USE will be listed.

   <output file>: Location of output html file. If not specified output will be
          written to the indivisual table files.

   type-list: comma-separated list of network object types or
              service object types

   Network Object Types: domain, gateway, group, host, logical,
                         machines_range, network, range, router, switch

   Service Object Types: group, icmp, other, ports_range, rpc, tcp, udp\n";
}
