#!/usr/bin/perl

#############################################################################
# fwrules.pl: Analyze and print a report of Firewall-1 rules and objects
#
# Copyright (C) 2001 Volker Tanger
# Copyright (C) 1999 Sean O'Neill
# Copyright (C) 1998 Stefan Moser
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# This program is based on previous work by an anonymous programmer who
# posted it to the FireWall-1 mailing list. Stephan Moser modified the program
# considerably and put it under the GPL.  He has moved on and Sean O'Neill took
# over the maintaince of this script. Later (Jan.2001) Volker Tanger picked up
# this job on Sean's suggestion.
#
# For bug reports and suggestions or if you just want to talk to me please
# contact me at volker.tanger@wyae.de
#
# Updates will be available at  http://www.wyae.de/software/
#
# Hopefully one day:
#     - Add logic to process user db definitions
#     - Add Wei Xu's (V.5.0) switch "--all_objs" which lists all objects regardless use.
#
# KNOWN ISSUES:  All the graphic objects capable of being used within FW-1 have
#                not been fully tested as my firewall ruleset does not utilize
#                all the objects.  If you have an object in your ruleset that
#                doesn't display correctly, send me updated code or a simple
#                ruleset and object file showing the problem.
#
#                RESSOURCES NOT SUPPORTED (that's where I am working on - Volker)
#
#                Members of Service or Network Object are not colored correctly.
#                This will be a bit more difficult to achieve as one will have to
#                change the current design of that routine.
#
#
# January 2001
# v4.2 to v6.0
#     - Volker Tanger
#     - slightly renamed - and skipping the (re-structured and broken) version 5.0
#       Maybe this starts a forking but I'll try to re-work (at least some
#       of) the features of Wei Xu's version 5.0 without breaking the program
#     - corrected colouring scheme (added tags in Policy) - came "for free" with
#       the changed logic.
#     - Solved these "Hopefully one day" issues:
#         * Add a switch to turn on / off the addition of the graphic icons.
#                 I just had to add and/or correct the doc / USAGE info
#         * Add switch to specify were to get graphic icons from instead of
#           hard coding path in script
#                 Done, use the   --icon-path=  parameter
#         * Change the logic to print the graphic icons.  I don't like all the
#           if statements ... but it works for now.
#                 Done that, too - with which the following comes more or less
#                 "for free":
#         * Add a switch to turn on / off the addition of the name links.
#                 Done, use the   --without_links  switch.
#
# October 1999
# v 4.1 to v4.2
#     - Gianluca Rotoni
#     - Fixed 'Negated objects and services' bug. The error occurred when a
#       network object or a service was negated in a rule (operator "not in").
#       The rule was parsed all right and appeared in the output but the negated
#       object or service didn't (unless the object/service was used somewhere
#       else in the rulebase in a non-negated form).
#
#     - Sean O'Neill
#     - Fixed four bugs.
#       1) Short log objects were not being handled at all.
#       2) Disabled rules were not being handled correctly. The rule would
#          print out OK but the HTML page wouldn't indicate that the rule was
#          actually disabled.  The word DISABLED (hyphenated) is printed
#          beneath the rule row number.
#       3) Fixed a problem with how the script handled rules based on Session
#          Authentication.  Noted by John Allen.
#       4) The grahpic icon for rules with "Reject" as the action were not
#          being handled correctly.
#
# July 10 1999 - Sean O'Neill
# v4.0 to v4.1
#     - Fixed a few bugs I missed
#     - Added logic to print &nbsp; character in the Address Translation Rules
#       section if nothing was printed in a cell.
#     - Added logic to print &nbsp; character in the Security Servers
#       section if nothing was printed in a cell for the various Welcome
#       Messages.
#
# July 1999 - Sean O'Neill - AppNet, Inc. - soneill@appnet.com
# Stefan Moser has indicated he will no longer be maintaining this script.  He
# is moving on.
# Modifications include:
#     - Added graphical icons to ruleset table.  Makes the display look more
#       like the real thing.
#     - Modified the script logic flow some.  This was done to enable the
#       addition of the graphic icons.  Basically, the objects are read in
#       before the rules are.
#     - Added href's around the graphic icon in the rules section to the object
#       definitions in the appropriate sections.
#     - Incorporated John Schoonover, Dave Long, and my updates together
#       Because of the coloring used in my firewall definitions, I made the
#       addition of the color tags off by default.  See --color in help for
#       help on this.
#
#     - NOTE that all icons types were not added.  My firewall doesn't have
#       everything defined but it has the basics defined.
#
# From fwrules3.pl:
# June 2, 1999 - Some slight but useful modifications to Stefan Moser's
# masterful work:
#      I have added html color tags to network and service objects to reflect
#          the colors used in the FW-1 GUI. Thanks to Doug Jacobson
#          (jacobson@pop.c-com.net) for his color chart at
#          http://www.phoenix.net/~jacobson/pages/rgbhex.html
#      I have provided anchors and links between network and service objects
#          and references to them as members of groups. This only works if
#          the linked items are in the same file as the anchors.
#      john.schoonover@usa.net.
#
# From an email addressed to Stefan Moser:
# June 1999 - Dave Long - dave.long@unitedkingdom.ncr.com
# Modifications include:
#     - Added logic to make the script properly handle FW-1 v4.0 encryption
#       rules.
#############################################################################

#
# Used modules
#
use English;
use FileHandle;
use Getopt::Long;

#
# Defaults, Constants
#
$SCR_NAME="fw1rules";
$VERSION="6.0";
$TRUE = 1;
$FALSE = 0;

$DEFAULT_OBJ_FILE = 'objects.C';
$DEFAULT_ICON_PATH = 'icons';

@NWO = qw(host gateway router switch network domain logical group blackbox machines_range);
@SVO = qw(group icmp other rpc tcp udp ports_range);

$rules = 0;

%fieldIndices =
  ('src'      => 1,
   'dst'      => 2,
   'services' => 3,
   'action'   => 4,
   'track'    => 5,
   'time'     => 6,
   'install'  => 7,
   'comments' => 8,
   'Encrypt'  => 9,
   'disabled'  => 10);

%adtrFieldIndices =
  ('src_adtr'                 => 1,
   'dst_adtr'                 => 2,
   'services_adtr'            => 3,
   'src_adtr_translated'      => 4,
   'dst_adtr_translated'      => 5,
   'services_adtr_translated' => 6,
   'install'                  => 7,
   'comments'                 => 8);

%colors =
  ('black'              => '<font color="#111111">',
   'blue'               => '<font color="#3333FF">',
   'cyan'               => '<font color="#00FFFF">',
   'dark green'         => '<font color="#006600">',
   'dark orchid'        => '<font color="#CC33CC">',
   'firebrick'          => '<font color="#993333">',
   'foreground'         => '<font color="#111111">',
   'forest green'       => '<font color="#66FF99">',
   'gold'               => '<font color="#FFCC33">',
   'green'              => '<font color="#00FF00">',
   'magenta'            => '<font color="#FF99FF">',
   'medium slate blue'  => '<font color="#6666FF">',
   'navy blue'          => '<font color="#000077">',
   'orange'             => '<font color="#FF9900">',
   'red'                => '<font color="#FF0000">',
   'sienna'             => '<font color="#996600">',
   'yellow'             => '<font color="#FFFF00">');

@headers =
  ('RULE',
   'SOURCE',
   'DESTINATION',
   'SERVICES',
   'ACTION',
   'TRACK',
   'TIME',
   'INSTALL ON',
   'COMMENTS');

@adtrHeaders =
  ('RULE',
   'SOURCE',
   'DESTINATION',
   'SERVICE',
   'SOURCE',
   'DESTINATION',
   'SERVICE',
   'INSTALL ON',
   'COMMENTS');


# Parse and process options
if (!GetOptions(\%optctl, 'network-objects', 'service-objects', 'properties',
                'objects|o=s', 'rules|r=s', 'types-n|tn=s', 'types-s|ts=s', 'icon_path|p=s',
                'all', 'help', 'version', 'with_colors', 'without_icons', 'without_links',
                'all_objs')
    || keys(%optctl) == 0 || $optctl{help} == 1 || $optctl{version} == 1 || @ARGV > 1
    || (defined($optctl{'n-types'}) && !defined($optctl{'network-objects'}))
    || (defined($optctl{'s-types'}) && !defined($optctl{'service-objects'})))
{
        if ($optctl{version} == 1)
        {
                print STDERR "$SCR_NAME version $VERSION\n";
        } else {
                usage();
        }

        exit;
}

if (defined($optctl{all}))
{
        $optctl{'network-objects'} = $optctl{'service-objects'} = '';
        $optctl{properties} = 1;
}

# Parse network object types
if (defined($optctl{'network-objects'}))
{
        if (defined($optctl{'types-n'}))
        {
                @nwoTypes = split(/,/, lc($optctl{'types-n'}));
                foreach $type (@nwoTypes)
                {
                        if (!grep(/^$type$/, @NWO)) {
                        die "Error: invalid network object type specified: $type\n";
                        }
                }
    }
        else
        {
                @nwoTypes = @NWO;
        }
}

# Parse service object type
if (defined($optctl{'service-objects'}))
{
        if (defined($optctl{'types-s'}))
        {
                @svoTypes = split(/,/, lc($optctl{'types-s'}));
                foreach $type (@svoTypes)
                {
                        if (!grep(/^$type$/, @SVO))
                        {
                                die "Error: invalid service object type specified: $type\n";
                        }
                }
        }
        else
        {
                @svoTypes = @SVO;
        }
}

# optical sugar options
if (defined($optctl{'icon_path'}))
{
        $iconPathName = $optctl{'icon_path'};
} else {
        $iconPathName = $DEFAULT_ICON_PATH;
}

if (defined($optctl{'with_colors'}))
{
        $with_colors = $TRUE;
} else {
#        %colors = ();  # just to be REALLY sure, ok, just as idea for fututre versions maybe  ;-)
        $with_colors = $FALSE;
}

if (defined($optctl{'without_links'}))
{
        $without_links = $TRUE;
} else {
        $without_links = $FALSE;
}

if (defined($optctl{'without_icons'}))
{
        $without_icons = $TRUE;
} else {
        $without_icons = $FALSE;
}

# handling of objects and rules
if (defined($optctl{'all_objs'}))
{
        $all_objs = $TRUE;
} else {
        $all_objs = $FALSE;
}


# Open input and output files
if (defined($optctl{'rules'}))
{
        $rules = 1;

        if (!open(RULES, $optctl{'rules'}))
        {
                print STDERR "$SCR_NAME: Couldn't open rule file '$optctl{'rules'}'\n";
                exit 1;
        }
}

if (defined($optctl{'objects'}))
{
        $objectsFileName = $optctl{'objects'};
} else {
        $objectsFileName = $DEFAULT_OBJ_FILE;
}

if (!open(OBJECTS, $objectsFileName))
{
        print STDERR "$SCR_NAME: Couldn't open objects file '$objectsFileName'\n";
        exit 1;
}

if (@ARGV > 0)
{
        if (!open(STDOUT, ">".$ARGV[0]))
        {
                print STDERR "$SCR_NAME: Couldn't open output file $ARGV[0]\n";
                exit 1;
        }
}
#
# Read in rules
#
process_rules(\%usedNO, \%usedSO);
#
# Read in objects
#
if (defined($optctl{'network-objects'}) ||
        defined($optctl{'service-objects'}) ||
        defined($optctl{'properties'}))
{
        process_objects(\%objects);
}
#
# Begin print processing
#
print "<html>\n";
print "<title>FireWall-1 Rule Base and Object Definitions</title>\n";
print "<body>\n";
#
# Print rules
#
if (defined($optctl{'rules'}))
{
        print_rules(\%usedNO, \%usedSO);
}
#
# Print various objects
#
if (defined($optctl{'network-objects'}) ||
        defined($optctl{'service-objects'}) ||
        defined($optctl{'properties'}))
{
        #
        # complete the list of used nw objects by exploding members
        #
        if (defined($optctl{'rules'}))
        {
                get_used_objects(\%objects, \%usedNO, \%usedSO);
        }
        #
        # Page Header
        #
        print "<h2>FireWall-1 Object Definitions</h2>\n";
}

if (defined($optctl{'network-objects'}))
{
        # Print title
        print "<h3>Network Objects</h3>\n";

        # Open table
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Name</th><th>Type</th><th>Location</th>";
        print "<th>FW-1</th><th>IP Address</th>";
        print "<th>Netmask</th><th>Gateway</th><th>Members</th>";
        print "<th>Comment</th></tr>\n";

        $key_found = $FALSE;
        print_info(\%objects, 'netobj', @nwoTypes);

        # Close Table
        print "</table>\n";

        # Print title
        print "<h3>Address Translation Network Objects</h3>";

        # Open table
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Name</th><th>Type</th><th>First IP Address</th>";
        print "<th>Last IP Address</th><th>Comment</th></tr>\n";

        $key_found = $FALSE;
        print_info(\%objects, 'netobjadtr', 'all');

        # Close Table
        print "</table>\n";
}

if (defined($optctl{'service-objects'}))
{
        # Print title
        print "<h3>Service Objects</h3>\n";

        # Open table
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Name</th><th>Type</th><th>Port/<br>Program</th>";
        print "<th>S_Port from</th><th>S_Port to:</th>";
        print "<th>Match</th><th>Prolog</th>";
        print "<th>Members</th><th>Comment</th></tr>\n";

        $key_found = $FALSE;
        print_info(\%objects, 'servobj', @svoTypes);

        # Close Table
        print "</table>\n";

        # Print title
        print "<h3>Address Translation Service Objects</h3>\n";

        # Open table
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Name</th><th>Type</th><th>Protocol</th><th>First Port</th>";
        print "<th>Last Port</th><th>Comment</th></tr>\n";

        $key_found = $FALSE;
        print_info(\%objects, 'servobjadtr', 'all');

        # Close Table
        print "</table>\n";
}

if (defined($optctl{'properties'}))
{
        $key_found = $FALSE;
        $props = '';
        print_info(\%objects, 'props', 'all');
}

$nowString = localtime();

print "<p><hr>Generated: $nowString\n by $SCR_NAME v$VERSION";

# HTML footer
print "</body>\n";
print "</html>\n";

close(RULES);
close(OBJECTS);

sub process_objects
{
        my ($hash) = @_;

        while (<OBJECTS>)
        {
                chop;
                s/^\s+//;

                return if (/^\)$/);

                if (/^\:([^\s]*) \(([^\)]*)$/)
                {
                        # This branch recursively handles 'open' lines, i.e. lines
                        # that do not end in a ')'
                        # e.g. ': (CSFB-mgmt-server'

                        $key = $1; $key2 = $2;

                        #          if ($key2 eq '')
                        #            {
                        #              $key2 = $key;
                        #                }

                        if ($key eq '')
                        {
                                $key = $key2;
                        }
                        if ($key2 eq '')
                        {
                                $key2 = $key;
                        }

                        #$lcKey = lc($key2);
                        $lcKey = lc($key);
                        $$hash{$lcKey}[0] = $key2;
                        process_objects(\%{$$hash{$lcKey}[1]});
                }
                elsif (/^\:([^\s]*) (.*)$/)
                {
                        # This branch recursively handles 'closed' lines, i.e. lines
                        # that end in a ')'
                        # e.g. ':comments ("London Management Server")'

                        $key = $1; $value = $2;

                        if ($key eq '')
                        {
                                # Group members
                                $key = 'members';
                        }

                        $lcKey = lc($key);
                        $$hash{$lcKey}[0] = $key;

                        if ($$hash{$lcKey}[1] ne '')
                        {
                                $$hash{$lcKey}[1] .= ', ';
                        }

                        if ($value =~ /^\((.*)\)$/)
                        {
                                $value = $1;
                        }

                        if ($value =~ /^\"(.*)\"$/)
                        {
                                $value = $1;
                        }

                        $$hash{$lcKey}[1] .= $value;
                }
        }
}


sub get_used_objects
{
        my ($objects, $usedNO, $usedSO) = @_;

        $nwObjects = $$objects{'netobj'}[1];

        foreach $name (keys(%$nwObjects))
        {
            $NegatednwObjects = "!" . $$nwObjects{$name}[0];
            if ((defined($$usedNO{$$nwObjects{$name}[0]})) ||
                (defined($$usedNO{$NegatednwObjects})) || ($all_objs == $TRUE) )
            {
                if (defined($$nwObjects{$name}[1]{'members'}))
                        {
                                @members = split(',', $$nwObjects{$name}[1]{'members'}[1]);

                                foreach $member (@members)
                                {
                                    $member =~ s/^\s*(.*)/$1/;
                                    $usedNO{$member} = 1;
                                }
                        }
                }
        }

        $svObjects = $$objects{'servobj'}[1];

        foreach $name (keys(%$svObjects))
        {
            $NegatedsvObjects = "!" . $$svObjects{$name}[0];
            if ((defined($$usedSO{$$svObjects{$name}[0]})) ||
                (defined($$usedSO{$NegatedsvObjects})) || ($all_objs == $TRUE) )
                {
                        if (defined($$svObjects{$name}[1]{'members'}))
                        {
                                @members = split(',', $$svObjects{$name}[1]{'members'}[1]);
                                foreach $member (@members)
                                {
                                        $member =~ s/^\s*(.*)/$1/;
                                        $usedSO{$member} = 1;
                                }
                        }
                }
        }
}


sub print_info(\%$@)
{
        my ($hash, $key, @types) = @_;
        my ($k, $actType);

        if ($key_found == $TRUE && $key eq 'props')
        {
                print_properties($hash, $key);
        }
        else
        {
                foreach $k (sort keys(%$hash))
                {
                        if ($key_found == $TRUE)
                        {
                                {
                                        $actType = lc($$hash{$k}[1]{type}[1]);
                                        next if (!grep(/^$actType$/, @types));
                                }

                                $NegateHash="!"."$$hash{$k}[0]";

                                if ($key eq 'netobj')
                                {
                                    if (!$rules || defined($usedNO{$$hash{$k}[0]}) || defined($usedNO{$NegateHash}))
                                        {
                                                print_network_object($hash, $k);
                                        }
                                }
                                elsif ($key eq 'servobj')
                                {
                                        if (!$rules || defined($usedSO{$$hash{$k}[0]}) || defined($usedSO{$NegateHash}))
                                        {
                                                print_service_object($hash, $k);
                                        }
                                }
                                elsif ($key eq 'netobjadtr')
                                {
                                        if (!$rules || defined($usedNO{$$hash{$k}[0]}) || defined($usedNO{$NegateHash}))
                                        {
                                                print_adtr_network_object($hash, $k);
                                        }
                                }
                                elsif ($key eq 'servobjadtr')
                                {
                                        if (!$rules || defined($usedSO{$$hash{$k}[0]}) || defined($usedSO{$NegateHash}))
                                        {
                                                print_adtr_service_object($hash, $k);
                                        }
                                }
                        }
                        else
                        {
                                if ($$hash{$k}[0] eq $key)
                                {
                                        $key_found = $TRUE;
                                }
                                if (ref($$hash{$k}[1]))
                                {
                                        print_info($$hash{$k}[1], $key, @types);
                                }
                        }
                }
        }
}

sub print_network_object
{
        my($hash, $key) = @_;

        $pName = $pLocation = $pIpAddress = $pNetmask = $pGateway = '-';
        $pMembers = $pComment = '-';
        $pFwInstalled = '-';

        # Name
        # FW-1 colors are affixed to the object names.
        # Html anchors added so group members can refer back to the objects.
        if ( $without_links == $FALSE )
        {
            $pName = '<a NAME="' . $$hash{$key}[0] . '"></a>' ;
        } else {
            $pName = '' ;
        }

        if ( $with_colors == $TRUE )
        {
            $pName = $pName . $colors{lc($$hash{$key}[1]{color}[1])} . $$hash{$key}[0] . '</font>';
        } else {
            $pName = $pName . $$hash{$key}[0];
        }

        # Type
        $pType = lc($$hash{$key}[1]{type}[1]);

        return if (!grep(/^$pType$/, @nwoTypes));

        if ($pType eq 'host')
        {
                $pType = 'Host';
                $pFwInstalled = 'No';
        }
        elsif ($pType eq 'gateway')
        {
                $pType = 'Gateway';
        }
        elsif ($pType eq 'router')
        {
                $pType = 'Router';
                $pFwInstalled = 'No';
        }
        elsif ($pType eq 'switch')
        {
                $pType = 'Switch';
                $pFwInstalled = 'No';
        }
        elsif ($pType eq 'network')
        {
                $pType = 'Network';
        }
        elsif ($pType eq 'domain')
        {
                $pType = 'Domain';
        }
        elsif ($pType eq 'logical')
        {
                $pType = 'Logical';
        }
        elsif ($pType eq 'group')
        {
                $pType = 'Group';
        }
        elsif ($pType eq 'blackbox')
        {
                $pType = 'Blackbox';
        }

        # Location
        if (defined($$hash{$key}[1]{location}))
        {
                if (lc($$hash{$key}[1]{location}[1]) eq 'external')
                {
                        $pLocation = 'External';
                }
                elsif (lc($$hash{$key}[1]{location}[1]) eq 'internal')
                {
                $pLocation = 'Internal';
                }
        }

        # Firewall installed?
        if (defined($$hash{$key}[1]{firewall}))
        {
                if (lc($$hash{$key}[1]{firewall}[1]) eq 'installed')
                {
                        $pFwInstalled = 'Yes';
                }
        }

        # IP-Address
        if (defined($$hash{$key}[1]{ipaddr}))
        {
                $pIpAddress = $$hash{$key}[1]{ipaddr}[1];
        }

        # Netmask
        if (defined($$hash{$key}[1]{netmask}))
        {
                $pNetmask = $$hash{$key}[1]{netmask}[1];
        }

        # Gateway
        if (defined($$hash{$key}[1]{gateways}) &&
        $$hash{$key}[1]{gateways}[1] !~ /^\s*$/)
        {
                $pIpGateway = $$hash{$key}[1]{gateways}[1];
        }

        # Members
        if (defined($$hash{$key}[1]{members}))
        {
                $pMembers = $$hash{$key}[1]{members}[1];
                # Remove white space
                $pMembers =~ s/\s*//g;
                # Provide links to object definitions
                if ($without_links == $TRUE){
                        $pMembers =~ s/([^,]+)/\1/g;
                } else {
                        $pMembers =~ s/([^,]+)/\<a href\=\"\#\1\"\>\1\<\/a\>/g;
                }
                # Replace ',' with <br>
                $pMembers =~ s/,\s*/<br>/g;
        }

        # Comments
        if (defined($$hash{$key}[1]{comments}))
        {
                $pComments = $$hash{$key}[1]{comments}[1];
                if (!length($pComments))
                {
                        $pComments = '-';
                }
        }

        # Interfaces?
        foreach $k (sort keys(%{$$hash{$key}[1]}))
        {
                if (ref($$hash{$key}[1]{$k}[1]) &&
                defined($$hash{$key}[1]{$k}[1]{iffullname}))
                {
                        $pName .= '<br>'.$$hash{$key}[1]{$k}[1]{iffullname}[1];

                        if (defined($$hash{$key}[1]{$k}[1]{ipaddr}))
                        {
                        $pIpAddress .= '<br>'.$$hash{$key}[1]{$k}[1]{ipaddr}[1];
                        }

                        if (defined($$hash{$key}[1]{$k}[1]{netmask}))
                        {
                                $pNetmask .= '<br>' . $$hash{$key}[1]{$k}[1]{netmask}[1];
                        }
                }
        }

        print "<tr valign=top>\n";
        print "<td>${pName}</td>";
        print "<td>$pType</td>";
        print "<td>$pLocation</td>";
        print "<td>$pFwInstalled</td>";
        print "<td>$pIpAddress</td>";
        print "<td>$pNetmask</td>";
        print "<td>$pGateway</td>";
        print "<td>$pMembers</td>";
        print "<td>$pComments</td></tr>";
}

sub print_adtr_network_object
{
        my($hash, $key) = @_;

        $pComment = '-';

        # Name
        $pName = $$hash{$key}[0];

        # Type
        $pType = lc($$hash{$key}[1]{type}[1]);

        return if (!grep(/^$pType$/, @nwoTypes));

        if ($pType eq 'machines_range')
        {
                $pType = 'Range';
        }

        # IP-Address
        if (defined($$hash{$key}[1]{ipaddr_first}))
        {
                $pIpFirstAddress = $$hash{$key}[1]{ipaddr_first}[1];
        }

        # IP-Address
        if (defined($$hash{$key}[1]{ipaddr_last}))
        {
                $pIpLastAddress = $$hash{$key}[1]{ipaddr_last}[1];
        }

        # Comments
        if (defined($$hash{$key}[1]{comments}))
        {
                $pComments = $$hash{$key}[1]{comments}[1];
                if (!length($pComments))
                {
                        $pComments = '-';
                }
        }

        print "<tr valign=top>\n";
        print "<td>${pName}</td>";
        print "<td>$pType</td>";
        print "<td>$pIpFirstAddress</td>";
        print "<td>$pIpLastAddress</td>";
        print "<td>$pComments</td></tr>";
}

sub print_service_object
{
        my ($hash, $key) = @_;

        $pPort = $pFrom = $pTo = $pMatch = $pProlog = $pMembers = $pComments = '-';

        # Name
        # FW-1 colors are affixed to the object names.
        # Html anchors added so group members can refer back to the objects.
        if ( $without_links == $FALSE )
        {
            $pName = '<a NAME="' . $$hash{$key}[0] . '"></a>' ;
        } else {
            $pName = '' ;
        }

        if ( $with_colors == $TRUE )
        {
            $pName = $pName . $colors{lc($$hash{$key}[1]{color}[1])} . $$hash{$key}[0] . '</font>';
        } else {
            $pName = $pName . $$hash{$key}[0];
        }

        # Type
        $pType = lc($$hash{$key}[1]{type}[1]);

        return if (!grep(/^$pType$/, @svoTypes));

        # Port
        if (defined($$hash{$key}[1]{port}))
        {
                $pPort = $$hash{$key}[1]{port}[1];
                if (!length($pPort))
                {
                        $pPort = '-';
                }
        }

        # Source port range / from
        if (defined($$hash{$key}[1]{src_port_from}))
        {
                $pFrom = $$hash{$key}[1]{src_port_from}[1];

                if (!length($pFrom))
                {
                        $pFrom = '-';
                }
        }

        # Source port range / to
        if (defined($$hash{$key}[1]{src_port_to}))
        {
                $pTo = $$hash{$key}[1]{src_port_to}[1];

                if (!length($pTo))
                {
                        $pTo = '-';
                }
        }

        # Match expression
        if (defined($$hash{$key}[1]{'exp'}))
        {
                $pMatch = $$hash{$key}[1]{'exp'}[1];

                if (!length($pMatch))
                {
                        $pMatch = '-';
                }
        }

        # Prolog
        if (defined($$hash{$key}[1]{'prolog'}))
        {
                $pProlog = $$hash{$key}[1]{'prolog'}[1];

                if (!length($pProlog))
                {
                        $pProlog = '-';
                }
        }

        # Group members
        if (defined($$hash{$key}[1]{members}))
        {
                $pMembers = $$hash{$key}[1]{members}[1];
                # Remove white space
                $pMembers =~ s/\s*//g;
                # Provide links to object definitions
#                $pMembers =~ s/([^,]+)/\<a href\=\"\#\1\"\>\1\<\/a\>/g;
                if ($without_links == $TRUE){
                        $pMembers =~ s/([^,]+)/\1/g;
                } else {
                        $pMembers =~ s/([^,]+)/\<a href\=\"\#\1\"\>\1\<\/a\>/g;
                }
                # Replace ',' with <br>
                $pMembers =~ s/,\s*/<br>/g;

                if (!length($pMembers))
                {
                        $pMembers = '-';
                }
        }

        # Comments
        if (defined($$hash{$key}[1]{'comments'}))
        {
                $pComments = $$hash{$key}[1]{'comments'}[1];

                if (!length($pComments))
                {
                        $pComments = '-';
                }
        }

        print "<tr valign=top>\n";
        print "<td>${pName}</td>";
        print "<td>$pType</td>";
        print "<td>$pPort</td>";
        print "<td>$pFrom</td>";
        print "<td>$pTo</td>";
        print "<td>$pMatch</td>";
        print "<td>$pProlog</td>";
        print "<td>$pMembers</td>";
        print "<td>$pComments</td></tr>";
}

sub print_adtr_service_object
{
        my ($hash, $key) = @_;

        $pComment = '-';

        # Name
        $pName = $$hash{$key}[0];

        # Type
        $pType = lc($$hash{$key}[1]{type}[1]);

        return if (!grep(/^$pType$/, @svoTypes));

        if ($pType == 'port_Range')
        {
                $pType = 'Range';
        }

        # Protocol
        if (defined($$hash{$key}[1]{ports_range_type}))
        {
                $pProtocol = $$hash{$key}[1]{ports_range_type}[1];
        }

        # First Port
        if (defined($$hash{$key}[1]{port_first}))
        {
                $pFirstPort = $$hash{$key}[1]{port_first}[1];
        }

        # Last Port
        if (defined($$hash{$key}[1]{port_last}))
        {
                $pLastPort = $$hash{$key}[1]{port_last}[1];
        }

        # Comments
        if (defined($$hash{$key}[1]{'comments'}))
        {
                $pComments = $$hash{$key}[1]{'comments'}[1];

                if (!length($pComments))
                {
                        $pComments = '-';
                }
        }

        print "<tr valign=top>\n";
        print "<td>${pName}</td>";
        print "<td>$pType</td>";
        print "<td>$pProtocol</td>";
        print "<td>$pFirstPort</td>";
        print "<td>$pLastPort</td>";
        print "<td>$pComments</td></tr>";
}

sub get_logtrack_value
{
        my ($hash, $propName) = @_;

        if (defined($$hash{$propName}[1]{type}))
        {
                return($$hash{$propName}[1]{type}[1]);
        }
        else
        {
                return('none');
        }
}

sub print_properties
{
        my ($hash, $key) = @_;

        if (defined($$hash{auth}))
        {
                $pAuthFailureTrack = 'auth';
        }
        elsif (defined($$hash{authalert}))
        {
                $pAuthFailureTrack = 'alert';
        }
        else
        {
                $pAuthFailureTrack = 'none';
        }

        # Print title
        print "<h3>Properties</h3>\n";

        # Open table
        print "<h4>Security Policy</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th><th>Value</th></tr>\n";

        print "<tr valign=top><td>Apply Gateway Rules to Interface Direction</td>";
        print "<td>n/a</td><td>$$hash{gatewaydir}[1]</td></tr>\n";

        print "<tr valign=top><td>TCP Session Timeout (sec)</td>";
        print "<td>n/a</td><td>$$hash{tcptimeout}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept Firewall-1 Control Connections</td>";
        print "<td>$$hash{fw1enable}[1]</td><td>$$hash{fw1_enable_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept UDP Replies</td>";
        print "<td>$$hash{udpreply}[1]</td><td>$$hash{udpreply_p}[1]</td></tr>\n";

        print "<tr valign=top><td>UDP Reply Timeout (sec)</td>";
        print "<td>n/a</td><td>$$hash{udptimeout}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept Outgoing Packets</td>";
        print "<td>$$hash{outgoing}[1]</td><td>$$hash{outgoing_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable Decryption on Accept</td>";
        print "<td>$$hash{acceptdecrypt}[1]</td><td>n/a</td></tr>\n";

        print "<tr valign=top><td>Use FASTPATH</td>";
        print "<td>$$hash{enable_fastpath}[1]</td><td>n/a</td></tr>\n";

        print "<tr valign=top><td>Accept RIP</td>";
        print "<td>$$hash{rip}[1]</td><td>$$hash{rip_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept Domain Name Queries (UDP)</td>";
        print "<td>$$hash{domain_udp}[1]</td><td>$$hash{domain_udp_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept Domain Name Download (TCP)</td>";
        print "<td>$$hash{domain_tcp}[1]</td><td>$$hash{domain_tcp_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept ICMP</td>";
        print "<td>$$hash{icmpenable}[1]</td><td>$$hash{icmpenable_p}[1]</td></tr>\n";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Services</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th><th>Value</th></tr>\n";

        print "<tr valign=top><td>Enable FTP PASV Connections:</td>";
        print "<td>$$hash{ftppasv}[1]</td><td>$$hash{ftppasv_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable RSH/REXEC Reverse stderr Connections</td>";
        print "<td>$$hash{rshstderr}[1]</td><td>$$hash{rshstderr_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable RPC Control</td>";
        print "<td>$$hash{rpcenable}[1]</td><td>$$hash{rpcenable_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable Response of FTP Data Connections:</td>";
        print "<td>$$hash{ftpdata}[1]</td><td>$$hash{ftpdata_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable Real Audio Reverse Connections:</td>";
        print "<td>$$hash{raudioenable}[1]</td><td>$$hash{raudioenable_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable VDOLive Reverse Connections</td>";
        print "<td>$$hash{vdolivenable}[1]</td><td>$$hash{vdolivenable_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable CoolTalk Data Connections (UDP)</td>";
        print "<td>$$hash{cooltalkenable}[1]</td><td>n/a</td></tr>\n";

        print "<tr valign=top><td>Enable H.323 Control and Data Connections</td>";
        print "<td>$$hash{iphoneenable}[1]</td><td>n/a</td></tr>\n";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Log and Alert</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th></tr>\n";

        print "<tr valign=top><td>Excessive Log Grace Period (sec)</td>";
        print "<td>$$hash{loggrace}[1]</td></tr>\n";

        print "<tr valign=top><td>PopUp Alert Command</td>";
        print "<td>$$hash{alertcmd}[1]</td></tr>\n";

        print "<tr valign=top><td>Mail Alert Command</td>";
        print "<td>$$hash{mailcmd}[1]</td></tr>\n";

        print "<tr valign=top><td>SNMP Trap Alert Command</td>";
        print "<td>$$hash{snmptrapcmd}[1]</td></tr>\n";

        print "<tr valign=top><td>User Defined Alert Command</td>";
        print "<td>$$hash{useralertcmd}[1]</td></tr>\n";

        print "<tr valign=top><td>Anti Spoof Alert Command</td>";
        print "<td>$$hash{spoofalertcmd}[1]</td></tr>\n";

        print "<tr valign=top><td>User Authentication Alert Command</td>";
        print "<td>$$hash{userauthalertcmd}[1]</td></tr>\n";

        $ipOptsLog = get_logtrack_value($hash, 'ipoptslog');

        print "<tr valign=top><td>IP Options Drop Track</td>";
        print "<td>$ipOptsLog</td></tr>\n";

        print "<tr valign=top><td>Log Established TCP Packets</td>";
        print "<td>$$hash{log_established_tcp}[1]</td></tr>\n";

        print "<tr valign=top><td>Enable Active Connections</td>";
        print "<td>$$hash{liveconns}[1]</td></tr>\n";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Resolving</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th></tr>\n";

        print "<tr valign=top><td>Lookup Priorities</td>";
        print "<td>1. $$hash{resolver_1}[1]<br>2. $$hash{resolver_2}[1]<br>";
        print "3. $$hash{resolver_3}[1]<br>4. $$hash{resolver_4}[1]</td></tr>\n";

        print "<tr valign=top><td>BIND Timeout (sec)</td>";
        print "<td>$$hash{timeout}[1]</td></tr>\n";

        print "<tr valign=top><td>BIND Retries</td>";
        print "<td>$$hash{retries}[1]</td></tr>\n";

        print "<tr valign=top><td>Log Viewer Resolver Properties</td>";
        print "<td>Page Timeout (sec): $$hash{pagetimeout}[1]</td></tr>\n";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Security Servers</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th></tr>\n";

        print "<tr valign=top><td>Telnet Welcome Message File</td>";
        if ( $$hash{telnet_msg}[1] eq "" )
        {
                print "<td>&nbsp;</td></tr>\n";
        } else {
                print "<td>$$hash{telnet_msg}[1]</td></tr>\n";
        }

        print "<tr valign=top><td>Rlogin Welcome Message File</td>";
        if ( $$hash{rlogin_msg}[1] eq "" )
        {
                print "<td>&nbsp;</td></tr>\n";
        } else {
                print "<td>$$hash{rlogin_msg}[1]</td></tr>\n";
        }

        print "<tr valign=top><td>FTP Welcome Message File</td>";
        if ( $$hash{ftp_msg}[1] eq "" )
        {
                print "<td>&nbsp;</td></tr>\n";
        } else {
                print "<td>$$hash{ftp_msg}[1]</td></tr>\n";
        }

        print "<tr valign=top><td>Client Authentication Welcome Message File</td>";
        if ( $$hash{clnt_auth_msg}[1] eq "" )
        {
                print "<td>&nbsp;</td></tr>\n";
        } else {
                print "<td>$$hash{clnt_auth_msg}[1]</td></tr>\n";
        }

        print "<tr valign=top><td>HTTP Next Proxy</td>";
        print "<td>Host: $$hash{http_next_proxy_host}[1]<br>";
        print "Port: $$hash{http_next_proxy_port}[1]</td></tr>\n";

        if (defined($$hash{http_servers}))
        {
                $servers = $$hash{http_servers}[1];

                print "<tr valign=top><td>HTTP Servers</td><td>";

                foreach $srv (keys(%$servers))
                {
                        print "Name: $$servers{$srv}[0]<br>";
                        print "Host: $$servers{$srv}[1]{host}[1]<br>";
                        print "Port: $$servers{$srv}[1]{port}[1]<br>";
                        print "Null Server: $$servers{$srv}[1]{ers}[1]<br>";
                        print "Reauthentication: $$servers{$srv}[1]{reauth_requests}[1]<p>     ";
                }

                print "</td></tr>\n";
        }

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Authentication</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th></tr>\n";

        print "<tr valign=top><td>User Authentication Session Timeout (min)</td>";
        print "<td>$$hash{au_timeout}[1]</td></tr>\n";

        $pAuthFailureTrack = get_logtrack_value($hash, default_track);

        print "<tr valign=top><td>Authentication Failure Track</td>";
        print "<td>$pAuthFailureTrack</td></tr>\n";

        print "<tr valign=top><td>AXENT Pathways Defender Server Setup</td>";
        print "<td>Server IP: $$hash{snk_server_ip}[1]<br>";
        print "Agent ID: $$hash{snk_agent_id}[1]<br>";
        print "Agent Key: $$hash{snk_agent_key}[1]<br>";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>SYNDefender</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th></tr>\n";

        if ($$hash{fwsynatk_method}[1] == 0)
        {
                $pFwSynAtkMethod = 'None';
        }
        elsif ($$hash{fwsynatk_method}[1] == 1)
        {
                $pFwSynAtkMethod = 'SYN Relay';
        }
        elsif ($$hash{fwsynatk_method}[1] == 2)
        {
                $pFwSynAtkMethod = 'SYN Gateway';
        }
        elsif ($$hash{fwsynatk_method}[1] == 3)
        {
                $pFwSynAtkMethod = 'Passive SYN Gateway';
        }

        print "<tr valign=top><td>Method</td>";
        print "<td>$pFwSynAtkMethod</td></tr>\n";

        print "<tr valign=top><td>Timeout</td>";
        print "<td>$$hash{fwsynatk_timeout}[1]</td></tr>\n";

        print "<tr valign=top><td>Maximum Sessions</td>";
        print "<td>$$hash{fwsynatk_max}[1]</td></tr>\n";

        print "<tr valign=top><td>Display Warning Messages</td>";
        print "<td>$$hash{fwsynatk_warning}[1]</td></tr>\n";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Miscellaneous</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th></tr>\n";

        print "<tr valign=top><td>Load Agents Port</td>";
        print "<td>$$hash{load_service_port}[1]</td></tr>\n";

        print "<tr valign=top><td>Load Measurement Interval</td>";
        print "<td>$$hash{lbalanced_period_wakeup_sec}[1]</td></tr>\n";

        # Close table
        print "</table>\n";

        # Open table
        print "<h4>Access Lists</h4>";
        print '<table border=2 cellpadding=5 bgcolor="Silver">';
        print "<tr valign=top><th>Property</th><th>Setting</th><th>Value</th></tr>\n";

        print "<tr valign=top><td>Accept Established TCP Connections</td>";
        print "<td>$$hash{established_router}[1]</td><td>$$hash{established_router_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept RIP</td>";
        print "<td>$$hash{rip_router}[1]</td><td>$$hash{rip_router_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept Domain Name Queries (UDP)</td>";
        print "<td>$$hash{domain_udp_router}[1]</td><td>$$hash{domain_udp_router_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept Domain Name Download (TCP)</td>";
        print "<td>$$hash{domain_tcp_router}[1]</td><td>$$hash{domain_tcp_router_p}[1]</td></tr>\n";

        print "<tr valign=top><td>Accept ICMP</td>";
        print "<td>$$hash{icmpenable_router}[1]</td><td>$$hash{icmpenable_router_p}[1]</td></tr>\n";

        # Close table
        print "</table>\n";
}

sub process_rules
{
        my ($objects) = @_[0];
        my ($services) = @_[1];

        $mode = -1;
        $ruleMode = 0;
        $adtrRuleMode = 1;

        $ruleNr = 0;
        $adtrRuleNr = 0;

        $printFlag = 0;

        while (<RULES>)
        {
                chop;
                $line++;

                s/^\s+//;

                if (/^\(\"\#\#(.+)\"/)
                {
                        $policyName = $1;
                }
                elsif (/^\(/ || /^\)/)
                {
                        next;
                }
                elsif (/^\:rule /)
                {
                        $ruleNr++;
                        $ruleTable[$ruleNr][0][0] = $ruleNr;
                        $mode = $ruleMode;
                }
                elsif (/^\:rule_adtr /)
                {
                        $adtrRuleNr++;
                        $adtrRuleTable[$adtrRuleNr][0][0] = $adtrRuleNr;
                        $mode = $adtrRuleMode;
                }
                elsif ($mode == $ruleMode)
                {
                        if (/^\:op \(\"not in\"\)/)
                        {
                                $not = '!';

                                if ($entryNr > 0)
                                {
                                        for ($i = 0; $i < $entryNr; $i++)
                                        {
                                                $ruleTable[$ruleNr][$column][$i] =
                                                $not . $ruleTable[$ruleNr][$column][$i];
                                        }
                                }
                        }
                        elsif (/^\:(\w+) /)
                        {
                                $col = $1;

                                if (defined($fieldIndices{$col}))
                                {
                                        $column = $fieldIndices{$col};
                                }
                                else
                                {
                                        next;
                                }

                                $entryNr = 0;
                                $not = '';

                                if ($col eq 'disabled')
                                {
                                        /^\:disabled\s*\((.*)\)\n$/;
                                        $value = $1;

                                        $ruleTable[$ruleNr][$column][$entryNr++] = $value;
                                }

                                if ($col eq 'comments')
                                {
                                        /^\:comments\s*(.*)\n$/;
                                        $value = $1;

                                        if ($value ne '()')
                                        {
                                                $value =~ s/^\(\"(.*)\"\)/$1/;
                                                $value =~ s/;/<br>/g;
                                                $ruleTable[$ruleNr][$column][$entryNr++] = $value;
                                        }
                                }
                        }
                        elsif (/^\: ([^)]+)$/)
                        {
                                chomp($value = $1);

                                if ($value =~ /^\(/)
                                {
                                        $value =~ s/^\(\"?([^\"]+)\"?$/$1/;
                                        do
                                        {
                                                $_ = <RULES>;
                                        }
                                        #
                                        # This fixes the problem with Session Authentication rules.
                                        # I don't like this because the script is now more dependant
                                        # on the syntax of the rules file. SON - Oct 1999
                                        #
                                        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                        # Major problem, if you edit the *.W file with a DOS editor
                                        # which will cause endless-loop.   Volker - Jan 2001
                                        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                        #
                                        while ( ! /^\t\t\)$/ );
                                        #while (/^\:/);
                                }

                                $ruleTable[$ruleNr][$column][$entryNr++] = $not . $value;
                        }
                }
                elsif ($mode == $adtrRuleMode)
                {
                        if (/^\:(\w+) /)
                        {
                                $col = $1;

                                if ($col eq 'adtr_method')
                                {
                                        if ($adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] ne 'Any')
                                        {
                                                if (/static/)
                                                {
                                                        $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] =
                                                        'S:'.$adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1];
                                                }
                                                elsif (/hide/)
                                                {
                                                        $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] =
                                                        'H:'. $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1];
                                                }
                                                else
                                                {
                                                        $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1] =
                                                        'U:'. $adtrRuleTable[$adtrRuleNr][$column][$entryNr - 1];
                                                }
                                        }

                                        next;
                                }

                                if (defined($adtrFieldIndices{$col}))
                                {
                                        $column = $adtrFieldIndices{$col};
                                }
                                else
                                {
                                        next;
                                }

                                $entryNr = 0;

                                if ($col eq 'comments')
                                {
                                        /^\:comments\s*(.*)\n$/;
                                        $value = $1;

                                        if ($value ne '()')
                                        {
                                                $value =~ s/^\(\"(.*)\"\)/$1/;
                                                $value =~ s/;/<br>/g;
                                                $adtrRuleTable[$adtrRuleNr][$column][$entryNr++] = $value;
                                        }
                                }
                        }
                        elsif (/^\: (.+)/)
                        {
                                $value = $1;

                                if ($value =~ /^\(/)
                                {
                                        $value =~ s/^\(\"?([^\"]+)\"?$/$1/;
                                        do
                                        {
                                                $_ = <RULES>;
                                        }
                                        while (/^\:/);
                                }

                                $adtrRuleTable[$adtrRuleNr][$column][$entryNr++] = $value;
                        }
                }
        }
}

sub print_rule_object
{
#-----svc_call--------------
#&print_rule_object ( $object_textname,
#                   lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]),
#                   lc($objects{'netobj'}[1]{$textObject_lc}[1]{'location'}[1]),
#                   lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]),
#                   $negated,
#                   ( $row == 0 && defined($ruleTable[$line][$fieldIndices{'disabled'}]) ),
#                   $row
#
#);
        my ($objname)           = @_[0];
        my ($netobj)            = @_[1];
        my ($netobj_type)       = @_[1];
        my ($netobj_location)   = @_[2];
        my ($servobj_type)      = @_[3];
        my ($negatedFlag)       = @_[4];
        my ($disabled)          = @_[5];
        my ($column)            = @_[6];
        my ($netobj_color)      = @_[7];
        my ($servobj_color)     = @_[8];

        my ($graphics)          = lc("$objname");
        my ($internal)          = '';
        my ($url)               = '';
        my ($color)             = "$netobj_color$servobj_color"; # wooha - caution, ugly hack:
                                                                 # only one of both is set simultaneously

        # now print all the stuff: first the HREF (if applicable)
        if ( ( (($column>0)&& ($column<4))  || ($column==7) ) &&
             ($objname ne 'Any') && ($objname ne 'Gateways') && ($without_links == FALSE) ) {
                $url = "<a border=0 href=\"#$objname\">";
        }
        print "$url";
        # now build the icon name and print it
        # the icon's name is derived from type or name - plus some additional
        # information. So be carfeul when renaming the icon files!
        if ( ($without_icons == $FALSE) && ($column != 0) && ($column != 8) ){
                # check for service or object type
                if ( ($column == 1) || ($column == 2) ){
                        $graphics = $netobj_type;
                } elsif ($column == 3) {
                        $graphics = $servobj_type;
                }
                # this is not elegant but works: usergroup detection
                if ($graphics =~ /\@/) {
                        $graphics = 'usergroup';
                }
                # first remove spaces from the name
                $graphics =~ s/ //g;
                # here come the additional graphics flags: internal-location and negated
                if ( "$netobj_location$netobj_type" eq 'internalgateway' ){
                        $internal = 'internal';
                }
                if ( ($objname eq 'Any') || ($objname eq 'Gateways') ){
                        $graphics = lc($objname);
                }
                if ( $negatedFlag == 1 ){
                        $graphics = "negated$graphics";
                }
                print "<image src=\"$iconPathName/$graphics$internal.gif\">";
        }
        # print color tags if wanted
        $colorp = ($with_colors == $TRUE) && ( "$color" ne "" );
        if ($colorp == $TRUE) {
                print "$colors{$color}";
        }
        # print object name (whooray!)
        print "$objname";
        if ( $disabled ){
                print "<BR><BR>DIS-<BR>ABLED";
        }
        # close color tags if necessary
        if ($colorp == $TRUE) {
                print "</font>";
        }
        # and close the HREF if necessary
        if ( $url ne '' ){
                print "</a>";
        }

}

sub print_rules
{
        my ($objects) = @_[0];
        my ($services) = @_[1];

        my ($textObject_lc);
        my ($object_textname);
        my ($negated) = 0;
        my ($printFlag) = 0;

        # Page Header
        print "<h2>Firewall Policy: $policyName</h2>\n";

        # Open table
        print '<table border=1 cellpadding=5 bgcolor="Silver"> <tr valign=top>';

        $maxRows = $#headers;

        # Table headers
        for ($row = 0; $row <= $maxRows; $row++)
        {
                print "<th>$headers[$row]</th>\n";
        }

        print "</tr>\n";

        for ($line = 1; $line <= $ruleNr; $line++)
        {
                if (!defined($ruleTable[$line][$fieldIndices{'time'}][0]))
                {
                        $ruleTable[$line][$fieldIndices{'time'}][0] = 'Any';
                }

                print "<tr valign=top>";

                for ($row = 0; $row <= $maxRows; $row++)
                {
                        $x = $#{$ruleTable[$line][$row]} + 1;

                        print "<td>";

                        for ($entryNr = 0; $entryNr < $x; $entryNr++)
                        {
                                $printFlag = 1;
                                if ($entryNr > 0)
                                {
                                        print "<br>";
                                }
                                #
                                # Figure out icon to place in front of object text name
                                #
                                # Check on non-user defined objects first, basically reserved word
                                # things like 'Any', 'Gateways', etc.
                                #
                                $negated = 0;
                                $object_textname = $ruleTable[$line][$row][$entryNr];
                                #
                                if ( $object_textname =~ /^\!(.*)/ )
                                {
                                        $object_textname = $1;
                                        $negated = 1;
                                }
                                $textObject_lc = lc( $object_textname );

                                # this subroutine replaces all the if - elsif... lines
                                # and adds a better control over links and colors
                                &print_rule_object ( $object_textname,
                                                lc($objects{'netobj'}[1]{$textObject_lc}[1]{'type'}[1]),
                                                lc($objects{'netobj'}[1]{$textObject_lc}[1]{'location'}[1]),
                                                lc($objects{'servobj'}[1]{$textObject_lc}[1]{'type'}[1]),
                                                $negated,
                                                ( $row == 0 && defined($ruleTable[$line][$fieldIndices{'disabled'}]) ),
                                                $row,
                                                lc($objects{'netobj'}[1]{$textObject_lc}[1]{'color'}[1]),
                                                lc($objects{'servobj'}[1]{$textObject_lc}[1]{'color'}[1])
                                );

                                if (($row == 1) || ($row == 2)) # this network object is marked as used
                                {
                                        $$objects{$ruleTable[$line][$row][$entryNr]} = 1;
                                }
                                elsif ($row == 3)               # this SERVICE objects is marked as used
                                {
                                        $$services{$ruleTable[$line][$row][$entryNr]} = 1;
                                }
                        }
                        #
                        # If $printFlag is 0, then nothing got printed in this cell.
                        # Print a space to fill in the table cell.
                        #
                        if ($printFlag == 0)
                        {
                                print "&nbsp;";
                        }

                        print "</td>";
                        $printFlag = 0;
                }

                print "</tr>\n";
        }

        # Close table
        print "</table><p>\n";

        if ($adtrRuleNr)
        {
                # Title
                print "<h3>Address Translation Rules</h3>\n";

                # Open table
                print '<table border=1 cellpadding=5 bgcolor="Silver">';

                # Table headers
                print "<tr valign=top><th rowspan=2>RULE</th><th colspan=3>ORIGINAL PACKET</th>";
                print "<th</th><th colspan=3>TRANSLATED PACKET</th><th rowspan=2>INSTALL ON</th>";
                print "<th rowspan=2>COMMENT</th></tr>\n";

                print "<tr valign=top>";

                $maxRows = $#adtrHeaders;

                for ($row = 1; $row <= ($maxRows - 2); $row++)
                {
                                print "<th>$adtrHeaders[$row]</th>\n";
                }

                print "</tr>\n";
                #
                # Reset $printFlag to zero
                #
                $printFlag = 0;
                #
                for ($line = 1; $line <= $adtrRuleNr; $line++)
                {
                        print "<tr valign=top>";

                        for ($row = 0; $row <= $maxRows; $row++)
                        {
                                $x = $#{$adtrRuleTable[$line][$row]} + 1;

                                print "<td>";

                                for ($entryNr = 0; $entryNr < $x; $entryNr++)
                                {
                                        if ($entryNr > 0)
                                        {
                                                print "<br>";
                                        }

                                        $printFlag = 1;

                                        print "$adtrRuleTable[$line][$row][$entryNr]";

                                        if (($row == 1) || ($row == 2) || ($row == 4) || ($row == 5))
                                        {
                                                if ($adtrRuleTable[$line][$row][$entryNr] =~ /.?\:(.+)/)
                                                {
                                                        $adtrRuleTable[$line][$row][$entryNr] = $1;
                                                }

                                                $$objects{$adtrRuleTable[$line][$row][$entryNr]} = 1;
                                        }
                                        elsif (($row == 3) || ($row == 6))
                                        {
                                                $$services{$adtrRuleTable[$line][$row][$entryNr]} = 1;
                                        }
                                }

                                if ($printFlag == 0)
                                {
                                        print "&nbsp;";
                                }
                                print "</td>\n";
                                $printFlag = 0;
                        }

                        print "</tr>\n";
                }

                # Close table
                print "</table>\n";

                print "<br><font size=2>LEGEND:  Address Translation Methods: S: Static, H: Hide</font><p>\n";
        }
}

sub usage
{
  print STDERR
  "usage: $SCR_NAME { --all | --network-objects [--tn=<type-list>] |
                   --service-objects [--ts=<type-list>] | properties}
                   [--objects=<objects file>] [--all_objs]
                   [--rules=<rule file>]
                   [--with_colors] [--without_links]
                   [--without_icons] [--icon_path=<path>] [<output file>]
       $SCR_NAME --help
       $SCR_NAME --version

   --all: List all objects (network objects, service objects and properties).
          This is the default.

   --network-objects: Include network objects in the output. See also '--tn'.
          If a rule base is specified, only network objects that appear in
          that rule base will be included in the output.

   --properties: Include property settings in the output.

   --service-objects: Include service objects in the output. See also '--ts'.
          If a rule base is specified, only service objects that appear in
          that rule base will be included in the output.

   --objects=<objects file>: Location of FireWall-1 objects file. Default is
          'objects.C'.

   --all_objs: it will list ALL services and/or objects. By defailt only those
          IN USE will be listed.

   --tn=<type-list>: List of network object types that are included in the
          output. See also 'type-list' below.

   --ts=<type-list>: List of service object types that are included in the
          output. See also 'type-list' below.

   --rules=<rule file>: Location of FireWall-1 rule file. If specified,
          FireWall-1 rule base will be included in output.

   --with_colors: Add html color tags to network and service objects to reflect
          the colors used in the FW-1 GUI.  By default, this is OFF.

   --without_links: by default objects are linked so you can jump to its definition.
          With this switch you can turn this behaviour off.

   --without_icons: do not include icon graphics in HTML output.
          By default, icons are included.

   --icon_path=<path>: Location of the icon graphic files. Defaults to ./icon

   <output file>: Location of output file. If not specified output will be
          written to STDOUT.

   type-list: comma-separated list of network object types or
              service object types

   Network Object Types: domain, gateway, group, host, logical,
                         machines_range, network, range, router, switch

   Service Object Types: group, icmp, other, ports_range, rpc, tcp, udp\n";
}
